'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''

import sys
from itertools import chain, repeat
import logging
logger = logging.getLogger(__file__)
from Bio import SeqIO, Seq
import Bio



#############################
#
# Dictionaries/maps
#
#############################

"""
a list of the unicode character encodings for the DNA alphabet
note 65 is A, 67 is C, 71 is G, 84 is T.
"""

letterToBinary={ # Unicode UTF-8 byte codes for standard NA residues: ACGT NOTE: letterToBinaryNA
    65: 0,
    67: 1,
    71: 2,
    84: 3
}
binaryToLetter=['A', 'C', 'G', 'T'] # FIXME: binaryToLetterNA
standard_letters=set("ACTG")


"""
IUPAC support mappings for the k-mer counter (NOT USED IN THE ASSEMBLER)
"""
IUPAC_NA_DOUBLETS = {
    "R": ["A", "G"],
    "Y": ["C", "T"],
    "S": ["G", "C"],
    "W": ["A", "T"],
    "K": ["G", "T"],
    "M": ["A", "C"]
}
IUPAC_NA_DOUBLET_CHARS=set(IUPAC_NA_DOUBLETS.keys())
IUPAC_NA_TRIPLETS = {
    "B": ["C", "G", "T"],
    "D": ["A", "G", "T"],
    "H": ["A", "C", "T"],
    "V": ["A", "C", "G"]
}
IUPAC_NA_TRIPLET_CHARS=set(IUPAC_NA_TRIPLETS.keys())
# *extremely* necessary variable
n = "N"
#############################
#
# Lambdas
#
#############################


get_doublet_chars = lambda x: IUPAC_NA_DOUBLETS[x] # list
get_triplet_chars = lambda x: IUPAC_NA_TRIPLETS[x]
replace_char = lambda seq, x, y: seq.replace(x, y) # str

standard_letters=set("ACTG") # Backwards compatibility
# IUPAC_NA_TRIPLET_CHARACTERS
# IUPAC_NA_DOUBLE_CHARACTERS
# ALL_IUPAC_SYMBOLS
__permitted_NA_characters = IUPAC_NA_TRIPLET_CHARS.union(IUPAC_NA_DOUBLET_CHARS).union(standard_letters) # set("ACTGRYSWKMBDHV")
__permitted_NA_characters_with_N = __permitted_NA_characters.union(set(n))
__core_NA_letters=set("ACTG") 




#############################
#
# Class definition
#
#############################

class Kmers:
    """A wrapper class to pass variables through the multiprocessing pool. Methods in this class are (c) Matthew Ralston as above/throughout. That said, the k-mer bit-shifting trick
    
    :ivar k: The choice of k to shred with
    :ivar strand_specific: Include k-mers from forward strand only (TO BE DEPRECATED)
    :ivar verbose: print extra logging in the case of fasta files
    :ivar all_metadata: return extra metadata about sequence locations
    """
    def __init__(self, k, strand_specific:bool=True, verbose:bool=False, all_metadata:bool=False):
        """

        :param k: The choice of k to shred with
        :type k: int
        :param strand_specific: Include k-mers from forward strand only (TO BE DEPRECATED)
        :type strand_specific: bool
        :param verbose: Whether or not to print extra logging in the case of fasta and not in the case of fastq
        :type verbose: bool
        :param all_metadata: Whether or not to pass back extra metadata (TO BE DEPRECATED)
        :type all_metadata: bool


        """
        if type(k) is not int:
            raise TypeError("kmerdb.kmer.Kmers expects an int as its first positional argument")
        elif type(strand_specific) is not bool:
            raise TypeError("kmerdb.kmer.Kmers expects a bool as its second positional argument")
        elif type(verbose) is not bool:
            raise TypeError("kmerdb.kmer.Kmers expects a bool as its third positional argument")
        elif type(all_metadata) is not bool:
            raise TypeError("kmerdb.kmer.Kmers expects a bool as its fourth positional argument")
        self.k = k 
        self.strand_specific = strand_specific
        self.verbose = verbose
        self.all_metadata = all_metadata
        self.__all_permitted_NA_characters = __permitted_NA_characters_with_N
        self.__permitted_NA_characters_with_N = __permitted_NA_characters_with_N
        self.__permitted_extended_NA_characters = __permitted_NA_characters

    def validate_seqRecord_and_detect_IUPAC(self, seqRecord:Bio.SeqRecord.SeqRecord, is_fasta:bool=True, quiet_iupac_warning:bool=True):
        """
        Helper method for validating seqRecord and warnings for non-standard IUPAC residues.

        :param seqRecord: a BioPython SeqRecord object 
        :type seqRecord: Bio.SeqRecord.SeqRecord 
        :param is_fasta: are the inputs ALL .fasta?
        :type is_fasta: bool
        :param quiet_iupac_warning: verbosity parameter
        :type quiet_iupac_warning: bool
        :raise ValueError: if *non* IUPAC symbols are detected in the seqRecord object.
        :returns: a set of the letters detected, and length of the validated Bio.SeqRecord sequence
        :rtype: tuple


        """
        if type(quiet_iupac_warning) is not bool:
            raise TypeError("kmerdb.kmer.validate_seqRecord_and_detect_IUPAC expects keyword argument 'quiet_iupac_warning' to be a bool")
        elif not isinstance(seqRecord, Bio.SeqRecord.SeqRecord):
            raise TypeError("kmerdb.kmer.validate_seqRecord_and_detect_IUPAC expects a Bio.SeqRecord.SeqRecord object as its first positional argument")
        letters = set(seqRecord.seq) # This is ugly. Should really be explicitly cast to str
        seqlen = len(seqRecord.seq)  # `` Ditto. Should be explicitly cast to str

        
        if seqlen < self.k:
            logger.error("Offending sequence ID: {0}".format(seqRecord.id))
            raise ValueError("kmerdb.kmer.validate_seqRecord_and_detect_IUPAC expects that each input sequence is longer than k.")
        
        extended_iupac_symbols = list(letters.intersection(self.__permitted_chars) - standard_letters)
        has_N = __permitted_NA_characters_with_N 
        all_non_iupac_symbols = list(letters - self.__permitted_chars)
        
        if len(all_non_iupac_symbols) > 0:
            logger.warning("One or more unexpected characters in the {0} sequence".format("fasta" if self.is_fasta else "fastq"))
            logger.warning(list(letters - self.__permitted_chars))
            raise ValueError("Non-IUPAC symbols detected in the fasta file")
        elif len(extended_iupac_symbols) > 0: # FIXME: 
            if quiet_iupac_warning is False:
                logger.warning("Will completely refuse to include k-mers with 'N'")
                logger.warning("All counts for k-mers including N will be discarded")
                logger.warning("Other IUPAC symbols will be replaced with their respective pairs/triads")
                logger.warning("And a count will be given to each, rather than a half count")
            elif quiet_iupac_warning is True:
                logger.warning("Suppressing warning that non-standard IUPAC residues (including N) are detected.")
        return (letters, seqlen)


    def _shred_for_graph(self, seqRecord:Bio.SeqRecord.SeqRecord):
        """
        Introduced in 0.7.7, required for valid assembly. Simply omits the rejection of IUPAC non-standard residues.

        :param seqRecord: a BioPython sequence object to shred into k-mers
        :type seqRecord: Bio.SeqRecord.SeqRecord
        :raise RuntimeError: Non-IUPAC and non-standard IUPAC (other than ATGC+N) residues detected
        :return: k-mer ids
        :rtype: list
        """
        kmers = []
        
        letters, seqlen = self.validate_seqRecord_and_detect_IUPAC(seqRecord, quiet_iupac_warning=True)


        # Iterate over *each* k-mer in the sequence by index
        for i in range(seqlen - self.k + 1):
            kmer_seq = seqRecord.seq[i:(i+self.k)] # Creates the k-mer as a slice of a seqRecord.seq
            # No non-standard IUPAC residues allowed for _shred for use in graph.py
            nonstandard_iupac_symbols = list(set(kmer_seq) - standard_letters)
            non_iupac_symbols = list(set(kmer_seq) - self.__permitted_chars)

            if len(non_iupac_symbols) > 1:
                logger.error("Non-IUPAC symbols:")
                logger.error(non_iupac_symbols)
                raise RuntimeError("Non-IUPAC symbol(s) detected...")
            elif len(nonstandard_iupac_symbols) == 0:
                #logger.debug("Perfect sequence content (ATCG) detected...")
                kmers.append(kmer_to_id(s))
            elif len(nonstandard_iupac_symbols) == 1 and iupac_symbols[0] == n:
                #logger.debug("Only N-abnormal sequence content detected (aside from ATCG)...")
                kmers.append(kmer_to_id(s))
            elif len(nonstandard_iupac_symbols) > 1:
                logger.error("Assembly cannot continue with this k-mer, and it will be discarded, possibly affecting assembly process")
                raise RuntimeError("Non-standard IUPAC symbols detected in .fa/.fq file...")
        return kmers
            
    def shred(self, seqRecord):
        """
        Take a seqRecord fasta/fastq object and slice according to the IUPAC charset.
        Doublets become replace with two counts, etc.

        :param seqRecord: 
        :type seqRecord: Bio.SeqRecord.SeqRecord
        :raise RuntimeError: No IUPAC characters detected or Non-IUPAC characters detected
        :raise ValueError: Non-IUPAC characters detected
        :returns: a dictionary of information about the Bio.SeqRecord shredded and the k-mers produced (including the IUPAC doublet/triplet expansion
        :rtype: dict

        """
        letters, seqlen = self.validate_seqRecord_and_detect_IUPAC(seqRecord, quiet_iupac_warning=True)

            
        kmers = []

        kmers = [seqRecord.seq[i:(i+self.k)] for i in range(seqlen - self.k + 1)]

        for s in kmers:
            iupac_symbols = list(set(list(s)).intersection(self.__permitted_chars))
            non_iupac_symbols = list(set(list(s)) - self.__permitted_chars)

            ######################################################################33
            #       I U P A C        m o d u l e
            ######################################################################33

            
            seqs = []

            if self.verbose is not False:
                sys.stderr.write("running IUPAC module. generating sequences. L252\n") # 10/1/24
                sys.stderr.write("system message: k-mer sequence -        [| {0} |]\n".format(s))
                sys.stderr.write("Iterating over k-mer for doublet/triplet iupac expansion\n")
            nonstandard_chars = (set(list(s)) - set(n)) - standard_letters
            for c in list(nonstandard_chars):

                print("Non-permitted character found: {0}".format(c))
                raise RuntimeError("can we get in here actually?")
                sys.stderr.write("character: {0}\n".format(c))
                sys.stderr.write("K-mer: {0}\n".format(s))
                sys.stderr.write("IUPAC symbols in the whole sequence: {0}\n".format(iupac_symbols))
                sys.stderr.write("Non-IUPAC symbols in the whole sequence: {0}\n".format(non_iupac_symbols))
                if len(iupac_symbols) > 0:
                    sys.stderr.write("expanding iupac doublets/triplets...\n")
                print("seq0: {0}".format(seqs))
                if c in standard_letters:
                    raise ValueError("Cannot have standard letters to this point")


                
                # print(c)
                # print("doublet expansions:")
                if c in IUPAC_TO_DOUBLET.keys():
                    for x in get_doublet_chars(c):
                        doublet1, doublet2 = replace_char(s, c, x)
                        seqs.append(doublet1)
                        seqs.append(doublet2)

                if c in IUPAC_TO_TRIPLET.keys():
                    for x in get_triplet_chars(c):
                        trip1, trip2, trip3 = replace_chars(s, c, x)
                        seqs.append(trip1)
                        seqs.append(trip2)
                        seqs.append(trip3)
            
            if seqs is None and len(non_iupac_symbols) > 0:
                raise ValueError("Non-IUPAC symbols detected in {0}".format(s))
            elif seqs is None and len(iupac_symbols) == 4 and len(set(iupac_symbols) - standard_letters) == 0:
                e = ValueError("standard iupac symbols likely. ATCG or a subset for this k-mer. Likely not generated from sequencing data")
                #raise e
                continue # 9/30/24 whooopsshuaazzzaaah

            elif seqs is None and len(iupac_symbols) > 0 and len(iupac_symbols) < 4:
                e = ValueError("just... stnadard iupac symbols likely. ATCG or a subset for this k-mer. Likely not generated from sequencing data") # 10/3/24
                #raise e
                continue
            # elif seqs is None and len(iupac_symbols) >0 and len(iupac_symbols) > 4:
            #     e = ValueError("extra IUPAC characters detected in {0}".format(s))
            #     #raise e

            #raise RuntimeError("Still debug. bona suarte")
            """
            You shall not pass! ... gas!
            Gandalyf the greatest
            """
            # if seqs is None:
            #     sys.stderr.write("K-mer Sequence: {0}\n".format(s))
            #     sys.stderr.write("IUPAC symbols found in sequence: {0}\n".format(iupac_symbols))
            #     sys.stderr.write("Non-IUPAC symbols found in sequence: {0}\n".format(non_iupac_symbols))
            #     raise RuntimeError("seqs array was not populated. Needs debugging...")

            # if seqs is None:
            #     raise RuntimeError("Cannot pass here without have seqs generated from shredding process")
            # else:
            #     print("Squeakuences")
            #     print("Seeeeqquueeencees:")
            #     print(seqs)


                
            for x in seqs:
                logger.debug(x)
                kmers.append(kmer_to_id(x))
                if not self.strand_specific: # Reverse complement by default
                    kmers.append(kmer_to_id(Seq.Seq(x).reverse_complement()))

                    # OKAY you are supposed to remember that
                    # the number of k-mer ids will no-longer match the appended metadata,
                    # So even though the lambda is correct, you have to do separate metadata for each id
            # elif len(non_iupac_symbols) > 0:
            #     sys.stderr.write("TEST CONDITIONS:\n")
            #     sys.stderr.write("Non-permitted characters should be 0: {0}".format(len(set(str(s)) - self.__permitted_chars)) + "\n")
            #     sys.stderr.write("IUPAC symbols should be > 0: {0}".format(len(iupac_symbols)) + "\n")
            #     sys.stderr.write("Non-IUPAC characters should be 0: {0}".format(len(non_iupac_symbols)) + "\n")
            #     sys.stderr.write("Non-IUPAC symbols: {0}".format(non_iupac_symbols) + "\n")
            #     sys.stderr.write("IUPAC symbols: {0}".format(iupac_symbols) + "\n")
            #     sys.stderr.write("K-mer: '{0}'".format(str(s)) + "\n")
            #     raise ValueError("uuuuuuuuuupppppppsssssssss")
            # else:
            #     try:
            #         kmers.append(kmer_to_id(s))
            #         if not self.strand_specific: # Reverse complement by default
            #             kmers.append(kmer_to_id(s.reverse_complement()))
            #     except KeyError as e:
            #         sys.stderr.write("One or more non-IUPAC letters found their way into a part of the code they're not supposed to go\n")
            #         sys.stderr.write("We officially support IUPAC but the statement challenging the sequence content failed, causing a genuine runtime error\n")
            #         sys.stderr.write("This caused the following KeyError\n")
            #         sys.stderr.write(e.__str__() + "\n")
            #         sys.stderr.write("Letters in the sequence: {0}".format(letters) + "\n")
            #         sys.stderr.write("Permitted letters: {0}".format(self.__permitted_chars) + "\n")
            #         raise RuntimeError("IUPAC standard extra base pairs (R, B, etc.) or non-IUPAC characters detected in the sequence")
            del s
        if self.verbose:
            sys.stderr.write("            --- ~~~ --- ~~~  shredded ~~~ --- ~~~ ---\n")
            sys.stderr.write("a {0}bp long sequence was shredded into L-k+1 {1} total and {2} unique k-mers\n\n{3} were discarded due to sequence content\n\n".format(len(seqRecord.seq), len(str(seqRecord.seq))-self.k+1, len(list(set(kmers))), len([x for x in kmers if x is None])) + "\n")




        # def kolmogorov_complexity(self, kmers:list):
        #     """

        #     """
        #     kmer_strings = list(map(lambda s: id_to_kmer(s), kmers))
        #     for i, s in enumerate(kmer_strings):


        #         letters = set(map(lambda c: s.count(c), standard_letters))

        #         match len(letters):
        #             case 0:
        #                 raise ValueError("Error during kmerdb.kmer.Kmer.shred.kolmogorov_complexity. Cannot calculate complexity of sequence with no letters")
        #             case 1:
        #                 return 1
        #             case _:
        #                 return _kolmogorov_complexity(s)
                              
        return {'id': seqRecord.id, 'kmers': kmers, 'read_length': seqlen}
            
        #return {'id': seqRecord.id, 'kmers': kmers, "seqids": repeat(seqRecord.id, len(starts)), "starts": starts, 'reverses': reverses}



#############################
#
# Functions
#
#############################

    

def kmer_to_id(s):
    """Convert a fixed length k-mer string to the binary encoding parameterized upon that same k

    Note that the conversion of a k-mer string to an id integer
    is consistent regardless of k, 
    because the k is implicit in the k-mer string's size.

    Therefore, this method does not need to be wrapped in the k-mer class

    Acknowledgements for the 'idx = idx << 2' bit-shifting trick goes to the authors of kPAL.

    @article{anvar2014determining,
    title={Determining the quality and complexity of next-generation sequencing data without a reference genome},
    author={Anvar, Seyed Yahya and Khachatryan, Lusine and Vermaat, Martijn and van Galen, Michiel and Pulyakhina, Irina and Ariyurek, Yavuz and Kraaijeveld, Ken and den Dunnen, Johan T and de Knijff, Peter and ’t Hoen, Peter AC and others},
  journal={Genome biology},
    volume={15},
    pages={1--15},
    year={2014},
    publisher={Springer}
    }


    :param s: The input k-mer as string
    :type s: str
    :raise TypeError: str argument required, non-str type detected
    :raise KeyError: Non-standard (ATCG) character detected
    :returns: The kPal-inspired binary encoding (thanks!)
    :rtype: int

    """

    if not isinstance(s, str) and type(s) is not str and not isinstance(s, Bio.Seq.Seq) and not isinstance(s, Bio.SeqRecord.SeqRecord): # Typecheck the input k-mer
        logger.error("Seq: {0}, type: {1}".format(s, type(s)))
        raise TypeError("kmerdb.kmer.kmer_to_id expects a str as its argument")
    elif s.find('N') != -1: # k-mer with 'N' do not have a binary encoding
        #logger.debug(TypeError("kdb.kmer.kmer_to_id expects the letters to contain only nucleotide symbols ATCG"))
        return None
    else: 
        idx = 0
        if isinstance(s, Bio.Seq.Seq) or isinstance(s, Bio.SeqRecord.SeqRecord):
            s = str(s)
        for c in bytes(s, "UTF-8"): # Use byteshifting for fast conversion to binary encoding
            idx = idx << 2
            try:
                idx = idx | letterToBinary[c]
            except KeyError as e:
                sys.stderr.write("Entire sequence: {0}".format(s) + "\n")
                sys.stderr.write("Problematic character: {0}".format(c) + "\n")
                raise e
        return idx


def id_to_kmer(id, k):
    """
    Convert an id_to_kmer. I don't understand this docstring's purpose.

    :param id: The int id is the input to the id_to_kmer conversion
    :type id: int
    :param k: The int k is used to byte convert the id to the sequence of characters
    :type k: int
    :raise TypeError: int argument required, non-int type detected
    :returns: The k-mer as string
    :rtype: str
    """
    if type(id) is not int:
        sys.stderr.write("kmer.id_to_kmer was given the following type as an id\n")
        sys.stderr.write(str(type(id)) + "\n")
        raise TypeError("kmerdb.id_to_kmer expects an int as its first positional argument")
    elif type(k) is not int:
        sys.stderr.write(str(type(k)) + "\n")
        raise TypeError("kmerdb.id_to_kmer expects an int as its second positional argument")
    else:
        kmer = ""
        for i in range(k):

            
            kmer += binaryToLetter[id & 0x03]
            id = id >> 2
        kmer = list(kmer)

        assert len(kmer) == k, "kmer.id_to_kmer encountered an inconsistency error"

        #just_reversed = kmer.reverse()
        kmer.reverse()
        return ''.join(kmer)


def neighbors(kmer:str, kmer_id:int,  k:int, quiet:bool=True):
    """

    3/11/24 revived. given a k-mer of length k, give its neighbors.

    This is so ugly.

    But rock on :
    

    :param kmer: The sequence as a string that will be sliced for k-mers
    :type kmer: str
    :param kmer_id: The k-mer id for neighbor generation
    :type kmer_id: int
    :param k: The int k 
    :type k: int
    :raise TypeError: Requires either a Bio.SeqRecord or str
    :raise TypeError: Requires k to be an int
    :raise ValueError: k must match the length of the input k-mer
    :returns: A list of 'neighboring' or 'adjacent' k-mers from derived from the input k-mer
    :rtype: list
    """
    if not isinstance(kmer, str):
        raise TypeError("kmerdb.kmer.neighbors expects a Biopython Seq object as its first positional argument")
    elif type(k) is not int:
        raise TypeError("kmerdb.kmer.neighbors expects an int as its second positional argument")
    elif len(kmer) != k:
        raise ValueError("kmerdb.kmer.neighbors cannot calculate the {0}-mer neighbors of a {1}-mer".format(k, len(s)))
    else:
        import copy
        letters1 = copy.deepcopy(binaryToLetter)
        letters2 = copy.deepcopy(letters1)
    
        firstCharRemoved = kmer[1:]
        lastCharRemoved  = kmer[:-1]

        new_type1 = list(map(lambda c: firstCharRemoved + c, letters1)) # Form the up neighbors

        new_type2 = list(map(lambda c: c + lastCharRemoved, letters2))  # Form the down neighbors
        
        """
        # TYPE 1: [first char removed ... + ... c  : ['A', "c", "g", "T"]]

        # TYPE 2: [["A", "C", "G", "T"] : c + last char removed  ]
        """
        new_type1_ids = list(map(kmer_to_id, new_type1))
        new_type2_ids = list(map(kmer_to_id, new_type2))

#         logger.debug(""" flower garden - joan G. Stark fir sur. fleicitaciones

#                                     wWWWw
#    vVVVv (___) wWWWw  wWWWw  (___)  vVVVv
#    (___)  ~Y~  (___)  vVVVv   ~Y~   (___)
#     ~Y~   \|    ~Y~   (___)    |/    ~Y~
#     \|   \ |/   \| /  \~Y~/   \|    \ |/
#    \\|// \\|// \\|/// \\|//  \\|// \\\|///
# jgs^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

#             """,

#             """   computing the neighbor structure uwu ...

#                   _(_)_                          wWWWw   _
#       @@@@       (_)@(_)   vVVVv     _     @@@@  (___) _(_)_
#      @@()@@ wWWWw  (_)\    (___)   _(_)_  @@()@@   Y  (_)@(_)
#       @@@@  (___)     `|/    Y    (_)@(_)  @@@@   \|/   (_)\
#        /      Y       \|    \|/    /(_)    \|      |/      |
#     \ |     \ |/       | / \ | /  \|/       |/    \|      \|/
# jgs \\|//   \\|///  \\\|//\\\|/// \|///  \\\|//  \\|//  \\\|// 
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# """,
#               "",

#         )

        
        if quiet is not True:
            sys.stderr.write("kmerdb.kmer.neighbors creating neighbors for k-mer {0} : '{1}'...".format(kmer_id, kmer) + "\n")
            sys.stderr.write(" ========================\n\n")
            sys.stderr.write(", ".join(list(map(str, new_type1_ids))) + "\n")
            sys.stderr.write(", ".join(list(map(str, new_type2_ids))) + "\n\n")
            sys.stderr.write(" ------------------------\n\n")

            sys.stderr.write("""
        k-id : {0}
        kmer : \"    {1}        \"

        'neighbors'

        {2}
        {3}

        'ids':

        {4}
        {5}
\n""".format(kmer_id, kmer, new_type1, new_type2, new_type1_ids, new_type2_ids))
        #return {"appended_first_char_all_ommitted": new_type1_ids, "prepended_last_char_all_omitted": new_type2_ids}


        
        return new_type1_ids + new_type2_ids



THANKS = """
==============
  Thanks
==============


Issues | https://github.com/MatthewRalston/kmerdb/issues










Github | https://github.com/MatthewRalston/kmerdb


Quix-start | https://matthewralston.github.io/kmerdb/quickstart


Pinned 5 - Issues :

--------------

1. " PINNED_TITLE_1 "

 :  Issue Text: """ + 5*" "  + """"\n" + ISSUE_TITLE
 
Summary :

SUMMARY_STR

- - - -
2. " Pinned Title #2 "

 :  Issue Text: """ + 5*" "  + """"\n" + PINNED_TITLE_2
 
Summary : 

PINNED_SUMMARY_STR_2

- - - -

3. " pinned title no. 3 "

 :  Issue Text: """ + 5*" "  + """"\n" + PINNED_TITLE_3
 
Summary :

PINNED_SUMMARY_STR_3

- - - -

4. " Pined titles numr 4 "

 :  Issue Text: """ + 5*" "  + """"\n" + PINNED_TITLE_4
 
Summary :

PINNED_SUMMARY_STR_4

- - - -


5. " Pinney the mooch "

 :  Issue Text: """ + 5*" "  + """"\n" + ISSUE_TITLE
 
Summary :

PINNED_SUMMARY_STR_5

- - - -





----
----


No downloads

NO_DOWNLOAD_ISSUE

----
----



"""



header = """

==========================================
               k m e r d b
==========================================

Thank you for using kmerdb. Please feel free
to submit issues and requests to 
https://github.com/MatthewRalston/kmerdb

"""
usage_note = "Use 'kmerdb [cmd] -h' for more details."


DONE = header + THANKS



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''




import sys
import os
import yaml, json


from collections import OrderedDict



from kmerdb import util


import logging
logger = logging.getLogger(__file__)


my_log_format = "%(levelname)s: %(asctime)s %(funcName)s L%(lineno)s| %(message)s"



def get_root_logger(level:int, logfile:str=None):
    if level is not None and type(level) is not int:
        raise TypeError("get_root_logger requires the log-level to be an int")

    elif logfile is not None and type(logfile) is not str:
        raise TypeError("kmerdb.logger.get_root_logger requires the logfile to be a str")
    elif os.path.exists(logfile):
        sys.stderr.write("Existing log-file '{0}' found. WARNING! Overwriting\n".format(logfile))
    
    levels=[logging.WARNING, logging.INFO, logging.DEBUG]
    if level < 0 or level > 2:
        raise TypeError("{0}.get_root_logger expects a verbosity between 0-2".format(__file__))
    logging.basicConfig(level=levels[level], format=my_log_format, datefmt="%Y/%m/%d %I:%M:%S")
    root_logger = logging.getLogger()

    
    for name in logging.Logger.manager.loggerDict.keys():
        if ('boto' in name) or ('urllib3' in name) or ('s3' in name) or ('findfont' in name):
            logging.getLogger(name).setLevel(logging.WARNING)

    
            
    return root_logger



class Loggah:
    
    def __init__(self, logfile:str=None, level:int=None):

        if logfile is None and type(logfile) is not str:
            raise TypeError("kmerdb.logger.Loggah() requires a str as its argument")

        elif level is None or type(level) is not int:
            raise TypeError("kmerdb.logger.Loggah() requires a int level")

        self._levels = ["DEBUG", "INFO", "WARNING", "ERROR"]
        self.logs = []
        self.logfile = logfile
        self.level = level
        self._logger = get_root_logger(self.level, logfile=logfile)


        if os.path.exists(logfile):
            sys.stderr.write("Overwriting existing log file '{0}'...\n".format(logfile))
            pass
        elif not os.path.exists(os.path.dirname(logfile)):
            sys.stderr.write("Creating new log file '{0}'...\n".format(logfile))

        if logfile is not None:
            
            fh = logging.FileHandler(logfile, mode="w")
            fh.setFormatter(my_log_format)
            fh.setLevel(logging.DEBUG)
            self._logger.addHandler(fh)
            
        
        
        
    def log_it(self, log_str:str, level:str="WARNING"):
        """
        NOTE: All formatting and type casting is out of scope of this function.


        This function *may* coerce simple dictionaries and lists to valid strings.


        The Python logging module outputs to stderr as well as to the log-file.
        """
        if level is not None and (type(level) is not str or level not in self._levels):
            raise TypeError("kmerdb.logger.Loggah.log_it: invalid log level '{0}'".format(level))
        elif log_str is not None and (type(log_str) is not str and type(log_str) is not dict and type(log_str) is not list):
            raise TypeError("kmerdb.logger.Loggah.log_it: Unhashable type '{0}' supplied to logging function. Cannot coerce to a loggable string...")

        # Try to coerce dict/list to str
        if type(log_str) is dict or type(log_str) is list:
            try:
                yaml.add_representer(OrderedDict, util.represent_yaml_from_collections_dot_OrderedDict)
                log_str = yaml.dump(log_str)
            except Exception as e:

                self.logger.error("Could not convert input to a loggable ASCII representation")
                self.logger.error(log_str)
                raise e
            

        if level == "DEBUG":
            self._logger.debug(log_str)
        elif level == "INFO":
            self._logger.info(log_str)
        elif level == "WARNING":
            self._logger.warning(log_str)
        elif level == "ERROR":
            self._logger.error(log_str)

        self.logs.append(log_str)



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''







import math
import numpy as np
import multiprocessing as mp
cimport numpy as cnp
cimport cython

import sys
#from numba import jit
#import functools

import logging
logger = logging.getLogger(__file__)



def d2_distance(a, b, sharedval):
    val = d2(a, b)
    sharedval.value = val

def d2(a: list, b: list):
    """
    simple D2 implementation
    """
    if type(a) is not list:
        raise ValueError("kmerdb.distance.d2 expects a list as its first positional argument")
    elif type(b) is not list:
        raise ValueError("kmerdb.distance.d2 expects a list as its second positional argument")
    l = len(a)

    if l != len(b):
        raise ValueError("kmerdb.distance.d2 vectors must be equal length")


    prod = 0
    for i in range(l):
        prod += a[i] * b[i]

    return prod



# cpdef long d2(int[:] a, int[:] b, bint presence_only):
#     cdef int d = 0

#     if len(a) != len(b):
#         raise ValueError("arrays must have equal size")
#     else:
#         for i in range(len(a)):
#             if a[i] == b[i]:
#                 d += 1
#             elif presence_only is True and a[i] > 0 and b[i] > 0:
#                 d += 1
#         return d


cpdef long lr(int[:] a,  int[:] ea):
    """
    The likelihood ratio statistic may be calculated from an array of word (k-tuple) abundances and their expectations.
    """
    cdef likelihood_ratio = 0

    if len(a) != len(ea):
        raise ValueError("arrays must have matching length")

    else:
        for i in range(len(a)):
            ratio = a[i]/ea[i]
            likelihood_ratio += a[i] * math.log2(ratio)

        return 2*likelihood_ratio



    
def pearson_correlation(a, b, total_kmers, sharedr):
    r = correlation(a, b, total_kmers)
    sharedr.value = r

#cpdef double correlation(cnp.uint64_t[:] a, cnp.uint64_t[:] b, int total_kmers):
cpdef double correlation(long[:] a, long[:] b, int total_kmers):

    cdef int i
    cdef long double ssxx = 0
    cdef long double ssyy = 0
    cdef long double ssxy = 0
    cdef long double xx = 0
    cdef long double yy = 0
    cdef long double xy = 0

    cdef float x_bar = np.sum(a)/total_kmers
    cdef float y_bar = np.sum(b)/total_kmers
    if total_kmers != len(a) or total_kmers != len(b):
        raise ValueError("NumPy kmer count array total does not match length of arrays")
    for i in range(total_kmers):
        try:
            xx = (a[i] - x_bar)**2
            yy = (b[i] - y_bar)**2
            xy = (a[i] - x_bar)*(b[i] - y_bar)
        except OverflowError as e:
            logger.error("{0}\t{1}   |   {2}\t{3}\t{4}".format(a[i], b[i], xx, yy, xy))
            logger.error(e)
            raise e
        

        ssxx += xx
        ssyy += yy
        if xy > 0: # Add a log of the negative of the negative number instead
            logger.debug("{0}\t{1}   |   {2}\t{3}\t{4} | ssxy = {5}".format(a[i], b[i], xx, yy, xy, ssxy))
            ssxy += xy
        else: # For large positive xy (not supposed to happen), we will interpret this as a inf/ssxx*ssyy

            logger.debug("{0}\t{1}   |   {2}\t{3}\t{4} | ssxy = {5}".format(a[i], b[i], xx, yy, xy, ssxy))
            ssxy += xy
            if ssxy < 0:
                logger.info("Sum of squared residuals is less than 0")
            elif ssxy == 0:
                logger.info("Sum of squared residuals is 0")
            elif ssxy > 0:
                logger.info("Calculating next ssxy...")
                continue
    logger.info("Custom Pearson correlation acquired")
    logger.info("{0}/sqrt({1}*{2})".format(ssxy, ssxx, ssyy))
    r = ssxy/(np.sqrt(ssxx*ssyy))
    return r
    



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''

import sys



import logging
logger = logging.getLogger(__file__)



import numpy as np
#cimport numpy as cnp
#cimport cython


#from kmerdb cimport cublass_gemm

#from cpython.ref cimport PyObject


# def linear_regression(a, b):
#     x = regression(a, b, method="numpy")

# @cython.boundscheck(False)
# @cython.wraparound(False)
# def regression(double[:, :] A, double[:] b, char[:] method):
#     """
#     Linear regression wth two inputs.
#     A 2D array 'A' with two dimensions, and non-necessarily square.

#     and a vector of data determining the linear constants in the regression.
    

#     least squares problems occur when b is not in the column space of A.
    
#     They are solvable if and only if the matrix A:
#       * is non-singular
#       * has independent rows/columns
#       * a non-zero determinant (i.e. it is *not* row identical to the identity matrix)
#       * and is full-rank. 

    
#     ############
#     # definition
#     ############
#     At(b-Ax) = 0 OR
#     AtAx = Atb

#     but, less developed...
#     Ax = b

#     These are the normal equations.
    
#     """
#     cdef int n, m
#     n = A.shape[0]  # number of samples
#     m = A.shape[1]  # number of constituents


#     if n != m:
#         raise ValueError("kmerdb.regression expects a square matrix A of independent variables for regression against the dependent variable.")

    
#     cdef double[:, :] At = np.empty((m, n), dtype=np.double)
#     cdef double[:, :] At_test = np.empty((m, n), dtype=np.double)

#     # AtA: square symmetric positive definite required next if A is non-singular.
#     cdef double[:, :] AtA = np.empty((m, m), dtype=np.double)

    
#     cdef double[:] Atb = np.empty(m, dtype=np.double)
#     cdef double[:] x = np.empty(m, dtype=np.double)

#     # Calculate A^T
#     #At = A.T
#     for i in range(n):
#         for j in range(m):
#             At[i, j] = A[j][i]
#     At_test = A.T
#     assert np.array_equal(At_test, At), "kmerdb.regression expects the transpose of A to be equivalent to NumPy's transpose function."

#     # Calculate A^T * A

#     ## [MULTIPLY]
#     if method == "numpy": # Delegated to BLAS if configured.
#         AtA = np.dot(At, A)
#     elif method == "strassen":
#         raise ValueError("NOTE: not yet implemented. testing a cublas delegated matrix multiply.")

#         #AtA = strassen_multiplication(At, A)
#     elif method == "gemm":

#         #cdef cnp.ndarray[int64_t, ndim=2] a = A
#         #cdef cnp.ndarray[int64_t, ndim=2] b = B
#         #cdef cnp.ndarray[int64_t, ndim=2] c = C



#         raise ValueError("NOTE: not implemented. issues during Cythonize. ")

    
        
#         #cublas_gemm.matrix_mult_cuda_int64(a, b, c) # Uses cublas_gemm, specifically Dgemm

        
#         #AtA = c
#         #AtA = np.ndarray(c, dtype="uint64")

#         raise ValueError("NOTE: not yet implemented. testing a cublas delegated matrix multiply.")
        
#     # Ensure the solution can be found by checking for non-singularity.

#     detAtA = np.linalg.det(AtA)

#     if detAtA == 0:
#         raise ValueError("kmerdb.regression encountered a non-invertible matrix 'AtA', which is used in the solution of the least-squares problem.")
#     else:
#         try:
#             # TODO: Need an inverse function 
#             AtA_inv = np.linalg.inv(AtA)
#         except Exception as e:
#             raise e


#     # Calculate A^T * b
#     Atb = np.dot(At, b)
    

#     # Inverse A^T A and left-multiply with Atb. This yields the estimates for x that minimize error from the projection of b, p on the left nullspace of AtA. At => multiply with b. x is the Euclidean/Newtonian minimum of the convex space spanned by the left 
#     x = AtA_inv.dot(Atb)

#     return x





'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''
# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False


import numpy as np
#from libc.stdlib cimport malloc, free
#from libc.string cimport memcpy

# A simple utility function to add two matrices
def add(A, B):
    i: cython.int
    j: cython.int
    rows: cython.int
    cols: cython.int

    rows = A.shape[0]
    cols = A.shape[1]
    
    C = np.empty((rows, cols), dtype=np.double)
    for i in range(rows):
        for j in range(cols):
            C[i, j] = A[i, j] + B[i, j]

    return C

# cdef cnp.ndarray[double, ndim=2] add(cnp.ndarray[double, ndim=2] A, 
#                                      cnp.ndarray[double, ndim=2] B):
#     cdef int i, j
#     cdef int rows = A.shape[0]
#     cdef int cols = A.shape[1]
#     cdef cnp.ndarray[double, ndim=2] C = np.empty((rows, cols), dtype=np.double)

#     for i in range(rows):
#         for j in range(cols):
#             C[i, j] = A[i, j] + B[i, j]
    
#     return C

# A simple utility function to subtract two matrices
# cdef cnp.ndarray[double, ndim=2] subtract(cnp.ndarray[double, ndim=2] A, 
#                                           cnp.ndarray[double, ndim=2] B):
#     cdef int i, j
#     cdef int rows = A.shape[0]
#     cdef int cols = A.shape[1]
#     cdef cnp.ndarray[double, ndim=2] C = np.empty((rows, cols), dtype=np.double)

#     for i in range(rows):
#         for j in range(cols):
#             C[i, j] = A[i, j] - B[i, j]
    
#     return C

# # Strassen's algorithm for matrix multiplication
# cdef cnp.ndarray[double, ndim=2] strassen(cnp.ndarray[double, ndim=2] A,
#                                           cnp.ndarray[double, ndim=2] B, str method):
#     cdef int n

    
#     n = A.shape[0]
#     if method is None:
#         method = "numpy"
    
    
#     if method == "numpy":
#         return np.dot(A, B)
#     if n <= 4:#: and base == "numpy"  # Base case
#         return np.dot(A, B)
#     elif n > 4 and method == "strassen":
    
#         # Splitting the matrices into quadrants
#         half = n // 2
#     cdef cnp.ndarray[double, ndim=2] A11 = A[:half, :half]
#     cdef cnp.ndarray[double, ndim=2] A12 = A[:half, half:]
#     cdef cnp.ndarray[double, ndim=2] A21 = A[half:, :half]
#     cdef cnp.ndarray[double, ndim=2] A22 = A[half:, half:]
    
#     cdef cnp.ndarray[double, ndim=2] B11 = B[:half, :half]
#     cdef cnp.ndarray[double, ndim=2] B12 = B[:half, half:]
#     cdef cnp.ndarray[double, ndim=2] B21 = B[half:, :half]
#     cdef cnp.ndarray[double, ndim=2] B22 = B[half:, half:]
    
#     cdef cnp.ndarray[double, ndim=2] P1
#     cdef cnp.ndarray[double, ndim=2] P2
#     cdef cnp.ndarray[double, ndim=2] P3
#     cdef cnp.ndarray[double, ndim=2] P4
#     cdef cnp.ndarray[double, ndim=2] P5
#     cdef cnp.ndarray[double, ndim=2] P6
#     cdef cnp.ndarray[double, ndim=2] P7
    
#     cdef cnp.ndarray[double, ndim=2] temp1 = B12
#     cdef cnp.ndarray[double, ndim=2] temp2 = A11
#     cdef cnp.ndarray[double, ndim=2] temp3 = A21
#     cdef cnp.ndarray[double, ndim=2] temp4 = B11
    
#     cdef cnp.ndarray[double, ndim=2] temp5 = A11
#     cdef cnp.ndarray[double, ndim=2] temp6 = B11
#     cdef cnp.ndarray[double, ndim=2] temp7 = A22
#     cdef cnp.ndarray[double, ndim=2] temp8 = B22
    
#     cdef cnp.ndarray[double, ndim=2] temp9 = A11
#     cdef cnp.ndarray[double, ndim=2] temp10 = B11
    
#     # Compute the 7 products
    
#     temp1 = subtract(B12, B22)
#     P1 = strassen(A11, temp1, method="numpy")
    
#     temp2 = add(A11, A12)
#     P2 = strassen(temp2, B22, method="numpy")
    
#     temp3 = add(A21, A22)
#     P3 = strassen(temp3, B11, method="numpy")
    
#     temp4 = subtract(B21, B11)
#     P4 = strassen(A22, temp4, method="numpy")
    
#     temp5 = add(A11, A22)
#     temp6 = add(B11, B22)
#     P5 = strassen(temp5, temp6, method="numpy")
    
#     temp7 = subtract(A12, A22)
#     temp8 = add(B21, B22)
#     P6 = strassen(temp7, temp8, method="numpy")
    
    
#     temp9 = subtract(A11, A21)
#     temp10 = add(B11, B12)
#     P7 = strassen(temp9, temp10, method="numpy")
    
#     # Combine the products into a single result matrix
#     C = np.empty((n, n), dtype=np.double)
    
#     # C1 =  ((P5 + P4) - P2) + P6) (2 additions, 1 subtraction times exactly proportional to the matrix sizes.

#     # temp1_2 = (P6 + (P2 + (P4 + P5))) #
#     # temp2 = (P6 + (P5+P4))
#     temp1_1 = np.add(P5, P4) # P5(A12 + A22  x B11 + B22)  ADD+  P4(A22 x B21 - B11)
#     temp1_2 = np.subtract(temp1, P2)
    
#     C1 = np.add(temp1_2, P6)
    

    
#     # t7 = (A12 - A22) # 9/20/24
#     # t31 = (B21 + B22) # 9/20
#     # Parts of p5. guess which operator combines
    
#     # C2 = (A11 + A22)  x (B11 + B22)

#     # temp2_1 = add(temp5, temp6) # 9/20
#     # should be multipliy, which makes these useless
#     # And it's just p5 then
#     temp2_3 = np.add(P5, P6) #  ((A12 - A22) x (B11 + B22)) + P6 (A12 - A22) x (B21 + B22)
    
    
    
    
#     C2 = np.add(P1, P2)
#     C3 = np.add(P3, P4)
    

        
        
        
        
#     C[:half, :half] = add(subtract(P1, P2), P6)
#     C[:half, half:] = add(P1, P2)
#     C[half:, :half] = add(P3, P4)
#     C[half:, half:] = subtract(add(subtract(add(P5, P1), P3), P7), P6)
        
#     return C

#     # End Strassen






'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''
import sys
import os

import math


#import yaml
#from collections import OrderedDict




import numpy as np

from kmerdb import kmer, config, util

#from kmerdb import logger as kdbLogger
#global logger = 




def select_lexicographical_minimizers(seq, seq_id, k, window_size, kmer_ids):
    """
    Select k-mers based on lexicographical minimizers and return a binary array indicating selected k-mers.
    
    :param seq: Input DNA sequence
    :type seq: str
    :param seq_id: Sequence ID as fasta header string
    :type seq_id: str
    :param k: Choice of k
    :type k: int
    
    :param window_size: Size of the sliding window
    :type window_size: int
    :param kmer_ids: Array of kmer IDs (1 to 4^k)
    :type kmer_ids: list

    :returns: Binary array where 1 indicates selected k-mer, 0 otherwise
    :rtype: int, np.array
    """

    
    
    N = len(kmer_ids)
    if not type(seq) is str:
        raise TypeError("kmerdb.minimizer.select_lexicographical_minimizers expects a str as its first positional argument")        
    if len(seq) < k:
        raise ValueError("Sequence length was less than k")
    
    sequence_length = len(seq)
    
    is_minimizer = np.zeros(sequence_length, dtype="uint32")

    L = N - k + 1


    coords = []

    for i in range(L): # Number of minimizers = floor(sequence_length/window_size)
        if i % window_size == 0:
            is_min = 1
            is_minimizer[i] = 1
            is_min = 1
        else:
            is_min = 0
            is_minimizer[i] = 0
            
                
        subseq = seq[i:i+k]
        kmer_id = kmer.kmer_to_id(subseq)

        coords.append((seq_id, i, kmer_id, is_min))
    
    # coords is [(seq_id, i:L, kmer_id, is_minimizer), ...]            
    return math.floor(L / window_size), is_minimizer, coords # coords is a list of kmer start positions in the input fasta sequence, referencing the kmer_id of the k-mer at the minimizer position, is_minimizer has dimension L = N-k+1, and the number of minimizers selected as first positional return, is_minimizer array as second positional return, and coordinate array with kmer_id values and 1:L/window_size number of elements pointing to the kmer_ids associated with the threetuple of (fasta_id, fasta_coord, kmer_id)



    # len(coord.subtract.keys)

def read_sequences_from_fastafile(fasta_file):
    """
    Parse sequence identifiers and sequences into two arrays.
    """
    fasta_seqs = []
    fasta_ids = []
    from Bio import SeqIO
    with open(fasta_file, mode="r") as ifile:
        for record in SeqIO.parse(ifile, "fasta"):
            fasta_seqs.append(str(record.seq))
            fasta_ids.append(str(record.id))
    return fasta_seqs, fasta_ids


def get_minimizers_from_kdbfile_and_input_fastafile(kdbfile:str, fasta_file:str, window_size, seqs:list=None, ids:list=None):


    from kmerdb import fileutil
    if seqs is None or ids is None:
        fasta_seqs, fasta_ids = read_sequences_from_fastafile(fasta_file)
    elif not all(type(fasta_id) is str for fasta_id, i in ids):
        raise TypeError("fasta_ids list is not a list of fasta sequence ids")
    elif not all(type(fasta_seq) is str for fasta_seq, j in seqs):
        raise ValueError("error no sequences")
    else:
        # Unimplemented
        fasta_seqs = seqs
        fasta_ids = ids

    coord_map = []
    coord_obj_array = []
    is_min = {}
        
    if fasta_file is None or type(fasta_file) is not str:
        raise TypeError("")
    elif type(kdbfile) is not str or not os.path.exists(kdbfile):
        raise ValueError("kmerdb.minimizer.get_minimizers_from_kdbfile_and_input_fastafile .kdb path does not exist on filesystem")
    else:
        with fileutil.open(kdbfile, mode='r', slurp=True) as kdb_in:
            metadata = kdb_in.metadata

            k = metadata["k"]
            kmer_ids_dtype = metadata["kmer_ids_dtype"]
            N = 4**metadata["k"]
            if metadata["version"] != config.VERSION:
                sys.stderr.write("KDB version is out of date, may be incompatible with current KDBReader class\n")
            kmer_ids = kdb_in.kmer_ids


            
            for i, seq_id in enumerate(fasta_ids):
                seq = fasta_seqs[i] 
                # print(seq)
                # print(seq_id)
                # print(i)

                # sys.exit(1)
                
                # is minimizer is a boolean array 0 or 1
                # coords is a numpy 32 array of kmer_ids
                sed_id = fasta_ids[i]
                seqlen, is_minimizer, coords = select_lexicographical_minimizers(seq, seq_id, metadata['k'], window_size, kmer_ids)

                is_min[seq_id] = is_minimizer
                """
                Fixme! coordmap should be primarily indexed on coordinate
                """
                coord_map = [[] for _ in range(seqlen)]
                #coord_map[kmer_ids[i]] = [] # coordmap becomes hashmap structure with kmer_id, fasta_seq_id1, start_coord1 , fasta_seq_id2, start_coord2

                
                for j in range(seqlen):
                    subseq = seq[j:j+k]
                    kmer_id = kmer.kmer_to_id(subseq)
                    coord_map.append({"fasta_id": fasta_ids[i], "sequence_length": seqlen, "coord": j, "kmer_id": kmer_id, "is_minimizer": is_minimizer[j]}) # Is_min should be the jth sequence element as a k-mer, and whether or not the jth k-mer is used as a minimizer


                    # fasta_id[i] where i is number of queries, len, the coord, and the is_minimizer bool


                    coord_obj_array.append((fasta_ids[i], seqlen, j, kmer_id, is_minimizer[j]))
                    # As hashmapped structure
                    #coord_obj_array[fasta_ids[i]].push((fasta_ids[i], seqlen, j, is_minimizer[j]))
                
                    
        return k, kmer_ids, fasta_ids, is_min, coord_map, coord_obj_array # kmer_ids are kmer actg ids, as read from the kdb file, fasta_ids is an array of inputs sequence ids, coordmap is a list of hashmaps


def print_coordmap(coordinate_mapping):
    """
    A hash of hashes, 2D associative array, not implemented further.


    
    
    """


    for fasta_id, sequence_length, basepair, is_min in coordinate_mapping:
        assert type(fasta_id) is str, "FASTA ID was not a sequence id string"
        assert type(sequence_length) is int, "Sequence length input was not a number"
        assert type(basepair) is int, " - error no coordinate \_( |>\">| )_/ ... ( ...wtf?)"
        assert type(is_min) is int, "should be typed numeric 1s and 0s"
        
        print("\t".join((fasta_id, sequence, basepair, is_min)))
    return






def minimizers_from_fasta_and_kdb(fasta_file, kdb_file, window_size):
    from Bio import SeqIO
    import numpy as np
    from kmerdb import fileutil, config, util, minimizer
    import json
    metadata = None
    N = None
    kmer_ids = None
    fasta_seqs = []
    fasta_ids = []
    mins = {}

    fa_sfx = os.path.splitext(fasta_file)[-1]
    kdb_sfx = os.path.splitext(kdb_file)[-1]

    
    if kdb_sfx != ".kdb" and kdb_sfx != ".kdbg": # A filepath with invalid suffix
        raise IOError("Input .kdb filepath '{0}' does not end in '.kdb'".format(kdb_file))
    elif not os.path.exists(kdb_file):
        raise IOError("Input .kdb filepath '{0}' does not exist on the filesystem".format(kdb_file))
    elif fa_sfx != ".fa" and fa_sfx != ".fna" and fa_sfx != ".fasta":
        raise IOError("Input .fasta filepath '{0}' does not exist on the filesystem".format(arguments.fasta))

    fasta_ids, fasta_seqs = read_sequences_from_fastafile(fasta_file)

    
            

    """
    Fixme
    """ 

            
    k, kmer_ids, _, is_minimizer, coordmap, coord_obj_array = get_minimizers_from_kdbfile_and_input_fastafile(kdb_file, fasta_file, window_size)

    return k, kmer_ids, fasta_ids, is_minimizer, coordmap, coord_obj_array # kmer_ids are kmer actg ids, as read from the kdb file, fasta_ids are a hashmap of inputs sequence ids, coordmap is a list of coordinate-as-index based minimizer locations. coordmap = [{"fasta_id": ..., "seqlen": n, }, ...] 



    
def read_minimizer_kdbi_index_file(kdbi1):
    """
    Read a minimizer file into memory
    """

    if kdbi_sfx != ".kdbi.1": # A filepath with invalid suffix
        raise IOError("Input .kdb index filepath '{0}' does not end in '.kdb'".format(arguments.kdbi1))

    minimizers = []
    with open(kdbi1, 'r') as minimizer_index_file:
        for l in minimizer_index_file:

            line = l.strip("\n").split("\t")

            assert len(line) == 2, "Index file corrupted. Use 'kmerdb minimizer' to regenerate the index file."
            kmer_id, is_minimizer = line
            
            if is_minimizer == 1:
                minimizers.append(kmer_id)
    return minimizers

def print_minimizer_kdbi_index_file(mins, kmer_ids, filename):
    """
    Print a minimizer array to a .kdbi.1 index file
    """

    if type(mins) is not list:
        raise TypeError("kmerdb.minimizer.print_minmimizer_kdbi_index_file epects its first positional argument to be a list")
    elif type(kmer_ids) is not list:
        raise TypeError("kmerdb.minimizer.print_minmimizer_kdbi_index_file epects its second positional argument to be a list")
    elif len(mins) != len(kmer_ids):
        raise ValueError("kmerdb.minimizer.print_minmimizer_kdbi_index_file expects the number of minimizers should match the number of potential kmers")
    elif type(filename) is not str:
        raise TypeError("kmerdb.minimizer.print_minmimizer_kdbi_index_file expects the third positional argument to be a str")
    
    with open(filename, 'w') as ofile:
        for kmer, i in enumerate(kmer_ids):
            ofile.write("{0}\t{1}\n".format(kmer, mins[i]))

    return filename
            



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''



VERSION="0.8.14" # LOL 03-01-25 omfg.
REQUIRES_PYTHON="3.7.4"
header_delimiter = "\n" + ("="*24) + "\n"


requirements_count = 11
requirements_dev_count = 8

# 9/22/24 Kind of? I mean...
# 4/9/24 there are some duplicates sure, but the requirements evolve faster than the dev do, are more essential for function, and I dont change the -dev file much. 

# 12 subcommands, and associated function names in __init__.py
subcommands = ["profile", "graph", "get_minimizers", "get_alignments", "matrix", "distance", "kmeans", "hierarchical", "view", "header", "index", "shuf", "usage", "help", "version"]
subcommand_functions = ["profile", "make_graph", "get_minimizers", "get_alignments", "get_matrix", "distances", "kmeans", "hierarchical", "view", "header", "index_file", "shuf", "expanded_help", "expanded_help", "version"]

default_logline_choices = (20, 50, 100, 200)
KDB_COLUMN_NUMBER = 4



graph_schema = {
    "type": "object",
    "properties": {
        "version": {"type": "string"},
        "metadata_blocks": {"type": "number"},
        "k": {"type": "number"},
        "total_kmers": {"type": "number"},
        "unique_kmers": {"type": "number"},
        "total_nullomers": {"type": "number"},
        "sorted": {"type": "boolean"},
        "n1_dtype": {"type": "string"},
        "n2_dtype": {"type": "string"},
        "weights_dtype": {"type": "string"},
        "tags": {
            "type": "array",
            "items": {"type": "string"}            
        },
        "files": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "filename": {"type": "string"},
                    "sha256": {
                        "type": "string",
                        "minLength": 64,
                        "maxLength": 64
                    },
                    "md5": {
                        "type": "string",
                        "minLength": 32,
                        "maxLength": 32
                    },
                    "total_reads": {"type": "number"},
                    "total_kmers": {"type": "number"},
                    "unique_kmers": {"type": "number"},
                    "nullomers": {"type": "number"}
                },
                "required": ["filename", "sha256", "md5", "total_reads", "total_kmers", "unique_kmers", "nullomers"]
            }
        },
        "comments": {
            "type": "array",
            "items": {"type": "string"}
        }
    },
    "required": ["version", "metadata_blocks", "k", "tags", "files", "total_kmers", "unique_kmers", "unique_nullomers", "n1_dtype", "n2_dtype", "weights_dtype"]
}

kdb_metadata_schema = {
    "type": "object",
    "properties": {
        "version": {"type": "string"},
        "metadata_blocks": {"type": "number"},
        "k": {"type": "number"},
        "total_kmers": {"type": "number"},
        "unique_kmers": {"type": "number"},
        "unique_nullomers": {"type": "number"},
        "kmer_coverage": {"type": "number"},
        "kmer_coverage_histogram_dtype": {"type": "string"},
        "total_nullomers": {"type": "number"},
        "metadata": {"type": "boolean"},
        "sorted": {"type": "boolean"},
        "profile_dtype": {"type": "string"},
        "kmer_ids_dtype": {"type": "string"},
        "count_dtypes": {"type": "string"},
        "frequencies_dtype": {"type": "string"},
        "dtype": {"type": "string"},
        "tags": {
            "type": "array",
            "items": {"type": "string"}
        },
        "files": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "filename": {"type": "string"},
                    "sha256": {
                        "type": "string",
                        "minLength": 64,
                        "maxLength": 64
                    },
                    "md5": {
                        "type": "string",
                        "minLength": 32,
                        "maxLength": 32
                    },
                    "total_reads": {"type": "number"},
                    "total_kmers": {"type": "number"},
                    "unique_kmers": {"type": "number"},
                    "nullomers": {"type": "number"}
                },
                "required": ["filename", "sha256", "md5", "total_reads", "total_kmers", "unique_kmers", "nullomers"]
            }
        },
        "comments": {
            "type": "array",
            "items": {"type": "string"}
        }
    },
    "required": ["version", "metadata_blocks", "total_kmers", "unique_kmers", "unique_nullomers", "k", "tags", "files", "kmer_ids_dtype", "profile_dtype", "frequencies_dtype", "count_dtype", "kmer_coverage", "kmer_coverage_histogram_dtype"]
}


exit_summary_schema = {
    "type": "object",
    "properties": {
        "subcommand": {"type": "string"},
        "kmerdb-version": {"type": "string"},
        "python-version": {"type": "string"},
        "feature": {"type": "number"},
        "feature_name": {"type": "string"},
        "feature_shortname": {"type": "string"},
        "feature_description": {"type": "string"},
        "step": {"type": "number"},
        "step_name": {"type": "string"},
        "step_shortname": {"type": "string"},
        "step_description": {"type": "string"},
        "log_file": {"type": "string"},
        "traceback": {"type": "string"},
        "error_file_name": {"type": "string"},
        "error_line_number": {"type": "number"},
        "error": {"type": "string"},        
    },
    "required": ["subcommand", "kmerdb-version", "python-version", "feature", "feature_name", "feature_shortname", "feature_description", "step", "step_name", "step_shortname", "step_description", "traceback", "error_file_name", "error_line_number", "error", ]
}






statfile =               "NAME.stats.txt"
logfile =                "NAME.stderr.log" # Primary logfile
output_usage_note =                "NAME.stdout.kmerdb.SUBCOMMAND.usage.txt"
#                "NAME.stderr.ascii.log",
primary_output =                "NAME.K.kdbg"


# FORMAT IN 
# kdb_program_metadata = {
#     "type": "object",
#     "properties": {
#         "outputs": {
#             "type": "array",
#             "items": [
#                 output_usage_note,
#                 logfile,
#                 statfile,
#                 primary_output,
#                 "NAME.K.kdb"
#             ]
#         },
#         "output_dir":
#         "logfile"
#     },
#     required = []
# }

# kdbg_program_metadata = {
#     "type": "object",
#     "properties": {
#         "outputs": {
#             "type": "array",
#             "items": [
#                 output_usage_note,
#                 logfile,
#                 statfile,
#                 primary_output,
#                 "NAME.K.kdbg"
#             ]
#         }
#         "output_dir": {
#             "type": "string"
#         }
#     },
#     required = ["output_dir", "outputs"]
# }
















pca_variance_fig_filepath = "PCA_variance_accumulation.png"
kmeans_elbow_graph_fig_filepath = "kmeans_elbow_graph.png"
kmeans_clustering_fig_filepath = "kmeans_clustering_of_kmer_profiles.png"
#ecopy_rarefaction_fig_filepath = "ecopy_rarefaction_curve.png"
hierarchical_clustering_dendrogram_fig_filepath = "dendrogram.png"
upgma_tree_phylip = "kdb_upgma_tree.phyloxml"
# files = (pca_variance_fig_filepath, kmeans_elbow_graph_fig_filepath, kmeans_clustering_fig_filepath, ecopy_rarefaction_fig_filepath, hierarchical_clustering_dendrogram_fig_filepath)

#######################################################

#          L o g o s

#######################################################



SPONGEBOB = """
⢀⣠⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣤⣄⡀
⣿⡋⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⠉⢙⣿
⣿⡇⠀⠀⠀⣠⣴⠾⠿⠷⣦⡀⢀⣴⠾⠿⠷⣦⣄⠀⠀⠀⢸⣿
⢸⡇⠀⠀⢰⡟⠁⠀⣀⣀⠈⢿⡿⠁⣀⣀⠀⠈⢻⡆⠀⠀⢸⡇
⢸⣷⠀⠀⢸⡇⠀⠀⠿⠿⠁⣸⣇⠈⠿⠿⠀⠀⢸⡇⠀⠀⣾⡇
⠘⣿⠀⠀⠈⠻⣦⣄⣀⣤⣾⠛⠛⣷⣤⣀⣠⣴⠟⠁⠀⠀⣿⠃
⠀⣿⠀⠘⢷⣄⠀⠉⠉⠙⢿⠄⠠⡿⠋⠉⠉⠀⣠⡾⠃⠀⣿⠀
⠀⣿⠀⠀⠀⠙⠻⢶⣦⣤⣤⣤⣤⣤⣤⣴⡶⠟⠋⠀⠀⠀⣿⠀
⠀⣿⡄⠀⠀⠀⠀⠀⣿⣀⣹⡇⢸⣏⣀⣿⠀⠀⠀⠀⠀⢠⣿⠀
⠀⢿⡇⠀⠀⠀⠀⠀⠉⠉⠉⠁⠈⠉⠉⠉⠀⠀⠀⠀⠀⢸⡿⠀
⠀⢸⣿⣿⣿⣿⣟⠛⠛⠛⣻⡿⢿⣟⠛⠛⠛⣻⣿⣿⣿⣿⡇⠀
⠀⢸⣿⣿⣿⣿⣿⣷⣤⣾⣿⣶⣶⣿⣷⣤⣾⣿⣿⣿⣿⣿⡇⠀
⠀⠘⣿⣿⣿⣿⣿⣿⣿⣿⣿⠃⠘⣿⣿⣿⣿⣿⣿⣿⣿⣿⠃⠀
⠀⠀⠉⠉⠉⠉⠉⠉⠉⠉⠻⣧⣼⠟⠉⠉⠉⠉⠉⠉⠉⠉⠀⠀
⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠈⠁⠀⠀⠀⠀⠀⠀⠀

"""




#######################################################

#          E n v i r o n m e n t

#######################################################




#SUBCOMMAND_NAME
#VERSION




# OUTPUTS determined at runtime
# OUTPUT_DIR/PREFIX determined at runtime
# Logfile (in output_dir) determined at runtime
# Verbosity determined at runtime


#######################################################

#          b a n n e r s  / mini - h e a d e r s

#######################################################


GITHUB_PROJECT_INFO = """
=======================================================
                  ||      G i t H u b     ||
=======================================================
                         Repo: kmerdb
                 Pinned issue: #130
               Feature branch: graph_algo
-------------------------------------------------------
"""







#######################################################

#          F o o t e r s

#######################################################


FOOTER_COMMAND_SUMMARY = """
command : COMMAND
runtime : RUNTIME
logfile : LOGFILE
exit_status : EXITCODE


0---------------------------0

.kdb_config file : CONFIG_PATH
was_pipeline : IS_PIPELINE

-----

conda_environment_validated? : CONDA_READY

"""









thanks = "\n\n\n" + "="*40 + """\n

  kmerdb v{0}    -     view -h|--help (or usage subcommand) for detailed information.


  please report any issues to https://github.com/MatthewRalston/kmerdb/issues and thanks.
""".format(VERSION)








#DONE = + usage + issue_tracker
# Accept the citation notice with 'kmerdb citation'
DONE = """
DONE\n
"""




































'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''



import sys
import os
import io
import gzip

import psutil
import tempfile

import yaml, json
from collections import deque, OrderedDict
import math
import re
from itertools import chain


from builtins import open as _open

import jsonschema
from Bio import SeqIO, Seq, bgzf


import numpy as np
#import networkx as nx

from kmerdb import fileutil, parse, kmer, config, util

# Logging configuration
global logger
logger = None



def open(filepath, mode="r", metadata=None, slurp:bool=False, logger=None):
    """
    Opens a file for reading or writing. Valid modes are 'xrwbt'. 
    Returns a lazy-loading KDBGReader object or a KDBGWriter object.
    The data may be force loaded with 'slurp=True'


    Opens a .kdbg file, still only view goals.

    Opens for reading or writing, legacy function.


    fn sig: (

    filepath
    mode(r/w param)
    metadata:None
    slurp: False

    )


    Stands in front of the kdb classes.

     also available in more general fileutil.py                                        

    metadata data vector (jsonschema)         described in config.py

    squash target unassigned on todo (slurp)


    nodes
    weights

    dtype defs, avail in json

    graph description avail in json

    traversal not implemented.


    


    :param filepath:
    :type filepath: str
    :param mode:
    :type mode: str
    :param metadata: The file header/metadata dictionary to write to the file.
    :type metadata: dict
    :param slurp: Immediately load all data into KDBGReader
    :type slurp: bool
    :raise ValueError: mode parameter related errors
    
    :returns: kmerdb.fileutil.KDBGReader/kmerdb.fileutil.KDBGWriter
    :rtype: kmerdb.fileutil.KDBGReader
    """
    if type(filepath) is not str:
        raise TypeError("kmerdb.graph.open expects a str as its first positional argument")
    # elif not os.access(filepath, os.R_OK):
    #     raise ValueError("kmerdb.graph.open expects an existing filepath as its first positional argument")
    elif type(mode) is not str:
        raise TypeError("kmerdb.graph.open expects the keyword argument 'mode' to be a str")
    elif (mode == "w" or mode == "x") and (metadata is not None and (isinstance(metadata, OrderedDict) or type(metadata) is dict)):
        pass
    elif (mode == "w" or mode == "x") and metadata is not None:
        raise TypeError("kmerdb.graph.open expects an additional metadata dictionary")
    elif type(slurp) is not bool:
        raise TypeError("kmerdb.graph.open expects a boolean for the keyword argument 'slurp'")
    modes = set(mode)
    if modes - set("xrwbt") or len(mode) > len(modes):
        raise ValueError("invalid mode: {}".format(mode))

    # oooo....


    # .......oooOO.ooo.....



    # ..........o.o.o.....O..


    # 
    creating = "x" in modes
    reading  = "r" in modes
    writing  = "w" in modes
    binary   = "b" in modes
    text     = "t" in modes

    if text and binary:
        raise ValueError("can't have text and binary mode at once")
    elif not (creating or reading or writing):
        raise ValueError("must have exactly one or read/write")

    if "r" in mode.lower():
        return KDBGReader(filename=filepath, mode=mode, slurp=slurp)
    elif "w" in mode.lower() or "x" in mode.lower():
        return KDBGWriter(filename=filepath, mode=mode, metadata=metadata)
    else:
        raise ValueError("Bad mode %r" % mode)


    self.logger = logger
    self._loggable = logger is not None


def bytesize_of_metadata(metadata):
    """
    defers to util.get_bytesize_of_metadata
    """
    return util.bytessize_of_metadata

    
def parse_kdbg_table_line(kdbg_table_line:str, sort_arr:bool=False, row_dtype:str="uint64"):
    """
    Parses a line according to the expected .kdbg syntax, and returns the python data types expected as a tuple.

    .kdbg


    table def
    =========================
    vector 'e'

    i          uint64
    node1_id         uint64
    node2_id         uint64
    w          uint64

    :param line:
    :type kdbg_table_line: str
    :raise TypeError: .kdbg table line was not str
    :raise TypeError: sort_arr was not bool
    :raise ValueError: .kdbg table has 4 columns - index, node1_id, node2_id, edge_weight
    :returns: node1_id, node2_id, weight
    :rtype: tuple
    """

    if type(kdbg_table_line) is not str:
        raise TypeError("kmerdb.graph.parse_kdbg_table_line expects a str as its first positional argument")
    elif type(kdbg_table_line) is str and kdbg_table_line == "":
        raise StopIteration("empty table line")
    
    if type(sort_arr) is not bool:
        raise TypeError("kmerdb.graph.parse_kdbg_table_line needs 'sort_arr' to be bool")
    try:
        np.dtype(row_dtype)
    except TypeError as e:
        sys.stderr.write("Invalid numpy dtype\n")
        raise e
    finally:



        linesplit = kdbg_table_line.rstrip().split("\t")
        
        if len(linesplit) != 4:
            sys.stderr.write("Full line:\n{0}".format(line) + "\n")
            raise ValueError("kmerdb.graph.parse_kdbg_table_line encountered a .kdbg line without 4 columns. Invalid format for edge list")
        else:
            i, node1_id, node2_id, weight = linesplit
            i, node1_id, node2_id, weight = int(i), int(node1_id), int(node2_id), int(weight)
        
            e = np.array([node1_id, node2_id, weight], dtype=row_dtype)
            
            # Still needs the sort_array default functionality as switchable.
            # Still needs the solver (ss-sp, or bfs)
            
            # # TODO: bfs would necessitate the objective
            # vanilla cpu implementation
            
            # # TODO:
            # Needs the node delexer.
            # is this a pre or post sort?
            # should be a post
            # Needs the signature and the implementation
            
            # Objective function def
            return [i] + list(e)
    

            

def parsefile(filepath:str, k:int, quiet:bool=True, b:int=50000, both_strands:bool=False, logger=None):
    """
    Parse a single sequence file.
    
    :param filepath: Path to a fasta or fastq file
    :type filepath: str
    :param k: Choice of k to shred k-mers with
    :type k: int
    :param b: Number of reads (per block) to process in parallel
    :type b: int
    :param both_strands: Strand specificity argument for k-mer shredding process
    :type both_strands: bool
    :raise TypeError: first argumentshould be a valid filepath
    :raise TypeError: second argument k should be an int
    :raise TypeError: quiet keyword arg should be a bool
    :raise TypeError: *unimplemented* - both_strands should be a bool
    :raise ValueError: Invalid length-of-0 kmer_id array produced during k-mer shredding
    :raise AssertionError: relationship 4^k = unique_kmers + (unique)nullomers found invalid
    :returns: (edge_list, header, counts, nullomer_array) header_dictionary is the file's metadata for the header block
    :rtype: (numpy.ndarray, dict, list, list)

    """

    """
    Type Checking

    """

    from kmerdb import kmer
    if type(filepath) is not str:
        raise TypeError("kmerdb.graph.parsefile expects a str as its first positional argument")
    elif not os.path.exists(filepath):
        raise OSError("kmerdb.graph.parsefile could not find the file '{0}' on the filesystem".format(filepath))
    elif type(k) is not int:
        raise TypeError("kmerdb.graph.parsefile expects an int as its second positional argument")
    elif type(b) is not int:
        raise TypeError("kmerdb.parse.parsefile expects the keyword argument 'b' to be an int")
    elif type(both_strands) is not bool:
        raise TypeError("kmerdb.graph.parsefile expects the keyword argument 'both_strands' to be a bool")

    _loggable = logger is not None
    

    N = 4**k
    counts = np.zeros(N, dtype="uint64")


    if _loggable:
        logger.log_it("Initializing edge list fileparser...", "DEBUG")
    seqprsr = parse.SeqParser(filepath, b, k)
    fasta = not seqprsr.fastq # Look inside the seqprsr object for the type of file
    
    # Initialize the kmer class for sequence shredding activities
    Kmer = kmer.Kmers(k) # A wrapper class to shred k-mers with
    
    # Actually load in records 'recs'
    recs = [r for r in seqprsr]


    if _loggable:
        logger.log_it("Read {0} sequences from the {1} seqparser object".format(len(recs), "fasta" if fasta else "fastq"), "INFO")
        
        
    while len(recs):
        num_recs = len(recs)
        #logger.debug("\n\nAcquiring list of all k-mer ids from {0} sequence records...\n\n".format(num_recs))


        list_of_dicts = list(map(Kmer._shred_for_graph, recs))

        if _loggable:
            logger.log_it("k-mer shredding a block of {0} reads/sequences".format(num_recs), "INFO")
        kmer_ids = list(chain.from_iterable(list_of_dicts))

        
        num_kmers = len(kmer_ids)

        if num_kmers == 0:
            raise ValueError("No k-mers to add. Something likely went wrong. Please report to the issue tracker")
        else:
            sys.stderr.write("\nAccumulating all k-mers from this set of records...\n")
            for kmer in kmer_ids:
                counts[kmer] += 1
        

        # END WHILE routine
        # load_more_records, according to 'block size'. Legacy syntax
        recs = [r for r in seqprsr] # The next block of exactly 'b' reads
        # This will be logged redundantly with the sys.stderr.write method calls at line 141 and 166 of parse.py (in the _next_fasta() and _next_fastq() methods)
        #sys.stderr("\n")
        # JUST LOGGING
        if _loggable:
            logger.log_it("Read {0} more records from the {1} seqparser object".format(len(recs), "fasta" if fasta else "fastq"), "DEBUG")




    if _loggable:
        logger.log_it("Read {0} k-mers from the input file".format(len(kmer_ids)), "INFO")

    sys.stderr.write("\n\n\n")
    if _loggable:
        logger.log_it("K-mer counts read from input(s)", "INFO")
    sys.stderr.write("="*40 + "\n")

    if _loggable:
        logger.log_it("Constructing weighted edge list...", "INFO")
    edge_list = make_graph(kmer_ids, k, quiet=quiet, logger=logger)

    if _loggable:
        logger.log_it("Number of edges: {0}".format(len(edge_list)), "INFO")
        logger.log_it("Edge list generation completed...", "INFO")

    unique_kmers = int(np.count_nonzero(counts))
    # FIXME
    num_nullomers = int(len(list(np.where(counts == 0)[0])))


    all_kmer_ids = list(range(N))
    is_nullomer = np.where(counts == 0)
    kmer_ids_np = np.array(all_kmer_ids, dtype="uint64")
    
    nullomer_array = list(kmer_ids_np[is_nullomer])
    
    assert num_nullomers == N - unique_kmers, "kmerdb.graph.parsefile found inconsistencies between two ways of counting nullomers. Internal error."

    
    seqprsr.total_kmers = int(np.sum(counts))
    seqprsr.unique_kmers = unique_kmers
    seqprsr.nullomers = num_nullomers
    seqprsr.nullomer_array = nullomer_array

    
    return (edge_list, seqprsr.header_dict(), counts, seqprsr.nullomer_array)

def make_graph(kmer_ids:list, k:int=None, quiet:bool=True, logger=None):
    """
    Make a neighbor graph from a k-mer id list.

    More specifically, this is a edge list

    Even more so, an edge list of technically adjacent k-mers in a .fa or .fq file.

    Taks from that what you will. It means its a 1D list with implicit, i.e. testable order, and no metadata.

    This particular adjacency list will have redundant relationships:

    i.e. they have been observed again elsewhere in the .fa sequence (1 or multiple) or the .fq file (millions of reads)

    So they are simply tallied.


    :param kmer_ids: list of k-mer ids
    :type kmer_ids: list
    :param k: Choice of k to shred k-mers with
    :type k: int
    :param quiet: Verbosity parameter
    :type quiet: bool
    :raise TypeError: first argument was not a list
    :raise TypeError: keyword argument k was not an inteer
    :raise TypeError: quiet keyword arg should be a bool
    :raise AssertionError: A *suppressed* error indicating whether two subsequent ids in the kmer_ids array were found to be unrelated by their sequences. Suppressed to support multi-sequence .fa/.fq files.
    :returns: An edge list, keyed on a k-mer id 2-tuple e.g. edges[(kmer_id1, kmer_id2)] = weight
    :rtype: dict


    """


    """
    DEBUG

    bad_pair

    A debugging parameter for assessing specific k-mer ids. The debugging process for self-assessing the k-mer pairs and the nature of the non-adjacency is not developed.
    
    3/18/24

    More specifically, i am still encountering the bug and haven't developed the algorithm beyond this point yet. 
    I'm still working on developing logging and processing throughout based on this debug variable set at this point in the variable namespace, and lexical order of the code.
    """

    bad_pair = (3050519, 3908357) # A debugging constant used throughout


    if kmer_ids is None or type(kmer_ids) is not list:
        raise TypeError("kmerdb.graph.make_graph expects a list as its only positional argument")
    elif k is None or type(k) is not int:
        raise TypeError("kmerdb.graph.make_graph expects the keyword argument k to be an int")
    elif quiet is None or type(quiet) is not bool:
        raise TypeError("kmerdb.graph.make_graph expects the keyword argument quiet to be a bool")

    _loggable = logger is not None
    # Initialize total_list of possible k-mer neighbor pairs from the edge list.
    # Step 1. Complete

    # # NIICE.
    # neighbors = None
    # #neighbors = 



    # arrangements[e] 


    # # this is an accumulator for lesser edges (edges where both count/frequency is exceedingly exceedingly rare.) air.air.
    
    # for e in edges_ids:
    #     if (arrangements[e] == 1):
    #         singletons[e] +=1
    #     elif (arrangements[e] == 0):
    #         nullomers[e] += 1
    #     elif (arrangements[e] == max(arrangements.values())):
    #         maxes = max(arrangements.values())

    #     all_neighbors[e] = 0

    # all_all_all_neighbors = None    
    # """
    # Iterate over all k-mer ids, creating lists of neighbors derived from the kmer.
    # Omit the self relationship, as the kmer is not a neighbor of itself.

    # 3/18/24

    # BACKLOG
    # TODO item
    # this isn't complete, as prepopulating the possible edges in k-space, is still in process among other things.
    # Convert to k-mer ids and add the adjacency tuple to the output

    # A minimal neighbor graph is a dictionary, keyed by the pair
    # """

    all_edges_in_kspace = {}


    """
    First we calc the possible neighbor space for the subspace of k-space observed in the file, and all of its 8 neighbors 
    obtained by taking off the leading character, and appending characters to the right of the k-1-mer.

    And vice versa.

    8*num_kmers_observed

    The first loop to populate that subspace p is simple enough that simplification or better than constant time would not be necessary.

    """



    """
    ["Actggtca"] = ["
    """
    neighbors = {}



    for i in range(len(kmer_ids)):

        k_id = kmer_ids[i]

               
        current_kmer = kmer.id_to_kmer(kmer_ids[i], k)
        common_seq = current_kmer[1:-1]

        local_neighbors = kmer.neighbors(current_kmer, k_id, k, quiet=quiet)
        
        neighbors[k_id] = local_neighbors


        # logger.error("             ----------------- 15633431 + 12202077 --------------- END ")


        # logger.error(" |||| =========     + = || = - = || = + = || = - = || + =          =========")

        # 3/18/24
        # LEGACY


        # marked as legacy
        
        # if k_id == 15633431 or k_id == 12202077:
        #     logger.debug("-"*11)
            
        #     logger.debug("K-mer {0} and all (non-self) neighbors:".format(k_id))
        #     logger.debug("{0}  -  {1}".format(current_kmer, "k1 + c"))
        #     logger.debug("-"*11)
        #     logger.debug("<:" + ",".join(char_first) + " |||| " + common_seq + "  ||||  " + ",".join(char_last) + ">:")
        #     logger.debug(",".join(list(map(str, char_first_ids))) + " |||| " + ",".join(list(map(str, char_last_ids))))


        #     logger.debug("Okay, so this is just the k-mer + either char first (left) or char last (right)")
        # for pair_ids in just_eight_edges:

        #     if 15633431 in pair_ids or 12202077 in pair_ids: # ACGT
        #         print(pair_ids)

            
        #     all_edges_in_kspace[pair_ids] = 0

        """
        Old win condition. I wanted to see that a neighbor pair that was originally not found:
        The 12-mers with the following ids, which had this subsequence in common 'GTGGATAACCT', was not found in the keys of the edge list dictionary.
        """

        # missing_key = (15633431, 12202077)
        # if missing_key in all_edges_in_kspace.keys():
        #     print("ITS THERE! LIAR")
        #     print("EDGE OF KSPACE")
        #     print("{0} : {1}".format(missing_key, all_edges_in_kspace[missing_key]))
        #     logger.error(" |||| =========     + = || = - = || = + = || = - = || + =          =========")
        
        # if (15633431, 12202077) in just_eight_edges:
        #     print("Sequence:")
        #     print(current_kmer)
        #     print(just_eight_edges)
            
        #     print("SUCCESS. This k-mer, current k-mer : {0}|{1} has 8 neighbors. {2} . None should be self".format(k_id, current_kmer, just_eight_edges))
        #     #print(all_edges_in_kspace)
        #     sys.exit(1)
        if _loggable:
            logger.log_it("adjacency space completed...", "INFO")
        pass

        
    """
    At this point, the adjacency structure is a simple list of tuples. Redundant tuples may exist in disparate parts of the graph.

    The structure of the neighbor_graph is [(from_node_id, to_node_id), ...]

    Next, the redundant k-mer id pairs are tallied, so that the edge weights are calculated
    """

    error_count = 0

    for i in range(len(kmer_ids)):
        if i+1 == len(kmer_ids):
            break
        else:
            """
            The following constitutes a 'neighboring' pair/edge of two k-mers (nodes):
            the k-mers are sequential in the original k-mer id list (implicitly sorted by virtue of the orientation of the read in the .fasta/.fastq file) 
            this list of k-mers actual neighbors can be validated by inspecting debug level output below (-vv).
            """
            # 3/18/24
            # premature optimization found. not gonna discuss.
            # 3/18/24
            # actually not? the point of the 'adjacency' in the simple list form from the KDBGReader invocation of the kmers structure from the /.fa/ format...
            # so its capturing the raw kmer array (of ids) from .fa and/or .fq data by virtue of some existing features
            # and it's collecting the proper ordering of the k-mer relationship integrity across the lines of the .fq
            # pair = (the current k-mer and its nearest neighbor in the sequence file i.e. whether or not it was adjacenct in the 'sequence' proper or whether the k-mer is in a list with multiple sequences, as in given from a .fq file, such that adjacent *reads* in the file produce a k-mer input list that has k-mer ids in a list that represents residue positions in both a .fa and a .fq file. Implicit order is understood by the programmer using this, but the computer itself still has no knowledge of whether a k-mer (or residue) in the file is adjacent in the genome coordinates to its literal neihbor in the list, vs on a subsequent chromosome, or Illumina read in the case of .fq input.
            """
            Basically at this point we haven't guaranteed anything about the input k-mer list. It's just been parsed and passed in from either a .fa or .fq file.
            Okay so now we're creating an artificial pair of k-mers, a relationship between two ids, that also constitutes a k-1-mer in common, regardless of the orientation of the 'neighbor' relationship. Additional assert statements below

            Essentially, this guarantees that, by having the k-2-mer in common, that the current_kmer and any of its 8 possible neighbors have at least a k-2mer in common among all types of neighbor relaionships.
            In addition, the 'neighbor' relationship between the current_kmer and a single neighbor, also has a k-1-mer in common.
            Initialize the k-mer pair (duple of ids) in the all_edges_in_kspace var, capturing the pair_id space in the keys iterable from the .keys()
            """

            current_kmer_id = kmer_ids[i]
            current_kmer = kmer.id_to_kmer(current_kmer_id, k)
            pair = (current_kmer, kmer.id_to_kmer(kmer_ids[i+1], k)) 
            pair_ids = tuple(map(kmer.kmer_to_id, pair))
            
            common_seq = current_kmer[1:]


            if quiet is False:
                sys.stderr.write("Kmer_id: = kmer_ids[i]     =        {0}\n".format(current_kmer_id))
                sys.stderr.write(" -----pair: = '{0}'  <-> '{1}'\n".format(pair[0], pair[1]))
            all_edges_in_kspace[pair_ids] = 0
            # Create the neighbor lists from the list of chars, prepended or appended to the k-mer suffix/prefix
            """
            BOGUS ASSERTIONS AND NONGUARANTEES...
            """
            assert len(current_kmer) == k
            assert len(common_seq) == k-1
            
            #logger.debug("Validations completed for k-mer {0}.".format(current_kmer_id))
            k1 = current_kmer[1:] # 11 characters (k - 1) remove the left most char, so new char goes on end to make the neighbor
            # not_the_king_but_the_rook = None
            k2 = current_kmer[:-1] # 11 characters ( k - 1) remove the final character so new char is prepended
            # Do we even assert these?

            # Comments:
            # The char_last requires its first char deleted.
            # Then a character is prepended to the beginning to the reverse of the sequence, or appending to the forward sequence.
            # The char_first requires its terminal char deleted, and then 4 residue characters are prepended to the forward sequence, or appended to the childless reverse
            # we call char_first 'childless' because it is assigned responsibility for the 4 interchangables on its prepend, while it's oldest parent is removed.
            """
            current_kmer
            pair : the pair of adjacent k-mers with the k-1mer as identical sequence
            pair_ids  : ids
            k1 : k-1mer   current_kmer[1:] left-most char ommitted, (neighbors *will* get appended to right side)
            k2 : k-1mer   current_kmer[:-1] right-most char ommitted, neighbors prepended

            # e.g. ACTGACTG
            # pair0 = CTGACTG (k-1 mer via first char removed. k-1 mer that receives appends)
            # pair1 = ACTGACT (k-1 mer via last char removed. k-1 mer that gets at its prepend.)

            DECLARATIONS FINISHED
            """
            # Commented 3/11/24 ish thanks nic
            # Working on some bills, then back to the drawing board.
            # I mean, aside from a lab job, if i could wfh or work localish then taking a month or so off to go to the mountains for soil sampling.
            # I'd really want to go with a guide or expert.
            # That way I could sample the most diverse ecosystems for sampling
            # But prior to that I'd need the go-kit prototype.
            # Maybe a soil sampling protocol, expert backpack, sampling fluids/solvents, portable lab cooler.
            # Then run the samples off of some grant.
            """
            LOCAL

            B A D         P A I R S
            debuging problematic pair of k-mer ids
            """
            if pair_ids == bad_pair:
                sys.stderr.write("Problematic pair of k-mer ids identified...\n")
                sys.stderr.write("Pair: ({0})\n".format(", ".join(pair_ids)))
                raise ValueError("DEBUGGING PURPOSES ONLY: Identified a pair of k-mers noted as problematic. If you see this error, file an issue at https://github.com/MatthewRalston/kmerdb")

            try:
                """
                This is real concise. performance focused, not too much style. 
                but this is the best way in python to concisely assert that the string slice is equivalent from k-mer to k-mer neighbor. Ensuring this throughout is damn routine.
                essentially, this code block is added as indication that these assertions hold on a single sequence .fa file
                by ensuring that sequential k-mer ids retrieved (via sliding window) from a single sequence .fa are, in fact, related by a common sequence.
                when the assertion error is triggered, it is most likely that this is due to subsequent reads in a .fastq file, resulting in an erroneous 'pair' of k-mers from the algorithm above, targeting the n'th and n+1'th k-mers in the kmer_id array.
                """
                assert pair[0][1:] == pair[1][:-1], "kmerdb.graph expects neighboring k-mers to have at least k-1 residues in common. Must be from a fastq file."
                assert (common_seq in pair[0]) and (common_seq in pair[1]), "kmerdb.graph expects the common_seq be in both sequences of the pair."
            except AssertionError as e:
                error_count += 1
                # logger.error(pair)
                # logger.warning("This k-mer pair should have this subsequence in common: {0}".format(common_seq))
                # logger.warning("k1: '{0}'           k2: '{1}'".format(k1, k2))
                # logger.warning("It is explicitly assumed this is from .fastq input")
                # #logger.debug("Types: pair[0] : {0} pair[1] : {1}".format(type(pair[0]), type(pair[1])))
                # logger.debug("'{0}' == '{1}'".format(pair[0][1:], pair[1][:-1]))
                # logger.debug("Is pair0[1:] == pair1[:-1]? : {0}".format(pair[0][1:] == pair[1][:-1]))
                # logger.debug(e)
                pass
            try:
                # Accumulate for each edge pair
                #if (15633431, 12202077) == pair_ids:
                if bad_pair == pair_ids:
                    sys.stderr.write("\n\nin accumulator... cannot identify the problem for this pairing again. It is in the wrong order in the key submission, and can't be recovered for some reason\n\n")
                    sys.stderr.write(bad_pair)

                    raise ValueError("Internal Error. Related to debugging problematic pairs during .fastq inter-read pair removal from edge list.")
                """
                ACCUMULATOR           --------- DONE
                """
                all_edges_in_kspace[pair_ids] += 1
                """

                You might be thinking "HOLY CRAP THATS A LOT OF MESSAGES" without the --quiet flag.

                That's right. Also:


                "Well, why can't you just fix the parser so it *only* returns valid edges from the .fq file?"

                i.e. check whether each k-mer... from each read... is at the final index of the read... right?


                So instead, I iterate over the k-mer array one time, and remove the edges. The performance difference between the approaches
                would be negligible. 

                """

                
                if quiet is False:
                    sys.stderr.write("Edge: {0} => {1}".format(pair[0], pair[1]))
#                     sys.stderr.write("""
#  ╱|、
# (˚ˎ 。7  
#  |、˜〵          
# じしˍ,)ノ
# \n""")                        
            except KeyError as e:
                sys.stderr.write("Invalid key(pair) used to access edge list\n")
                sys.stderr.write("PAIR: {0} => {1}\n".format(pair[0], pair[1]))
                sys.stderr.write("pair ids: {0}\n".format(pair_ids))
                sys.stderr.write(30*"=" + "\n")
                # for i, p in enumerate(all_edges_in_kspace.keys()):
                #     if bad_pair == pair_ids:
                #         # Double checking
                #         logger.error("Debugging the 'bad pair', problematic k-mer id pair...")
                #         logger.error(pair_ids)
                raise e
            # print("heeeeeey, wow thens doo the prin funshen")
            # print(pair_ids, pair, "wow thans prin funshun ( ---------)_______________________prin fun-shun_______________________________________(    --------------)")
            # print("print funshun you're so smart and tall")
            # print("print funshun why did you sit next to me")
            # print("function")


            maxlen = 15 # Placeholder
            p1str = str(pair_ids[0])
            p2str = str(pair_ids[1])
            p1str = p1str + (maxlen - len(p1str))*" "
            p2str = p2str + (maxlen - len(p2str))*" "

            if quiet is False:
                sys.stderr.write("k-mer 'pair' ({0}, {1}) adjacency from input file(s) verified".format(p1str, p2str))
                sys.stderr.write("\r")
            pass


    if error_count > 0:
        sys.stderr.write("\n\n\nNOTE: ADJACENCY ERRORS DETECTED: Found {0} 'improper' k-mer pairs/adjacencies from the input file(s),\n where subsequent k-mers in the k-mer id array (produced from the sliding window method over input seqs/reads) did not share k-1 residues in common.\n These *may* be introduced in the array from the last k-mer of one seq/read (in the .fa/.fq) and the first k-mer of the next seq/read.\n".format(error_count))
    sys.stderr.write("\n\n\n")
    sys.stderr.write("All edges populated *and* accumulated across k-mer id arrays from inputs.\n")
    sys.stderr.write("\n\n\n'Graph' generation complete\n\n")

            
    # This variable is unrestricted. A larger k may inflate this var beyond the memory limits of the computer where the observed k-space can be represented with enough resolution through enough profile diversity.
    """
    node_pairs and the k-space diversity 


    Depending on inputs, this may become large.
    """
    return all_edges_in_kspace



def create_graph(nodes:list, edge_tuples:list, gpu:bool=False):

    import networkx as nx

    if nodes is None or type(nodes) is not list or not all(type(n) is not int for n in nodes):
        raise TypeError("kmerdb.graph.create_graph expects the first argument to be a list of ints")
    elif edge_tuples is None or type(edge_tuples) is not list or not all(len(e) != 3 for e in edge_tuples) or not all((type(e[0]) is int and type(e[1]) is int) for e in edge_tuples):
        raise TypeError("kmerdb.graph.create_graph expects the second argument to be a list of tuples of length 2")
    elif gpu is None or type(gpu) is not bool:
        raise TypeError("kmerdb.graph.create_graph expects the keyword argument gpu to be a bool")


    """
    Now we make the networkx graph
    """
    G = nx.Graph()

    G.add_nodes_from(nodes)
    
        
    G.add_edges_from(edge_tuples)


    

    

def w_lexer():
    pass

            
def v_order_lexer(e:tuple, asc:bool=False):
    """
    Uh.
    """
    if type(asc) is not bool:
        raise TypeError("kmerdb.graph requires valid bool keyword arg 'asc'")
    
    if type(e) is not tuple:
        logger.error("kmerd.graph.threetuple_v_order_lexer needs a valid tuple")
        
        sys.exit(1)
        exit(1)
    elif len(e) != 2:
        raise IndexError("kmerdb.graph.v_order_lexer requires length 2")
    elif asc is not bool:
        logger.error("kmerdb.graph.v_order_lexer accepts the sort order as an input")
        raise TypeError("kmerdb.graph requires a lexical order")
    elif asc is False:
        raise RuntimeError("requires the order")

    
    if len(e) != 3:
        logger.error("kmerdb.graph invalid input")
        raise TypeError("kmerdb.graph requires a 3-tuple")

    n1, n2, w = e

    # Maybe...
    n1 = int(n1)
    n2 = int(n2)

    # This ... smh
    if asc is True:
        if n1 > n2:
            return (n2, n1, w)
        elif n2 > n1:
            return (n1, n2, w)
        else:
            raise RuntimeError('x')
    elif asc is False:
        if n1 > n2:
            return (n1, n2, w)
        elif n2 > n1:
            return (n2, n1, w)
        else:
            raise RuntimeError('x')
    else:
        
        logger.error("idgaf")
        return (n1, n2, w)


def randomsort_lexer(g):
    return (random.randint(0, 1428945), random.randint(0, 1428945), random.randint(0, 1428945))

def wrupsort_lexer(g):
    return sorted([g[0], g[1]])


    
def wrongsort_lexer(g, asc:bool=False):
    if asc is True:
        if n1 > n2:
            return (n1, n2, w)
        elif n2 > n1:
            return (n2, n1, w)
        else:
            raise RuntimeError('x')
    elif asc is False:
        if n1 > n2:
            return (n2, n1, w)
        elif n2 > n1:
            return (n1, n2, w)
        else:
            raise RuntimeError('x')




class KDBGReader(bgzf.BgzfReader):
    """
    A class to read .kdbg files.




    1 args

    8 keyword args

    fileobj
    mode
    n1_dtype
    n2_dtype
    weights_dtype
    sort
    slurp

    :ivar filename: The choice of k to shred with
    :ivar fileobj: An existing fileobject from io.IOBase
    :ivar mode: read/write mode
    :ivar n1_dtype: NumPy dtype
    :ivar n2_dtype: NumPy dtype
    :ivar weights_dtype: NumPy dtype
    :ivar sort: *unimplemented* - uh.. sort the .kdbg edges somehow
    :ivar slurp: Autoload the .kdbg file

    
    """
    def __init__(self, filename:str, fileobj:io.IOBase=None, mode:str="r", n1_dtype:str="uint64", n2_dtype:str="uint64", weights_dtype:str="uint64", sort:bool=False, slurp:bool=False, logger=None):
        """
        A wrapper around Bio.bgzf.BgzfReader

        :param filename: A valid filepath
        :type filename: str
        :param fileobj: An existing fileobject from io.IOBase
        :type fileobj: io.IOBase
        :param mode: read/write mode
        :type mode: str
        :param n1_dtype: NumPy dtype
        :type n1_dtype: str
        :param n2_dtype: NumPy dtype
        :type n2_dtype: str
        :param weights_dtype: NumPy dtype
        :type weights_dtype: str
        :param sort: *unimplemented* - uh.. sort the .kdbg edges somehow
        :type sort: bool
        :param slurp: Autoload the .kdbg file
        :type slurp: bool
        :raise TypeError: fileobj was not a file-type interface, inheriting from io.IOBase
        :raise TypeError: filename was not a str
        :

        """
        """

         fileobj type checking
        """


        if fileobj is not None and not isinstance(fileobj, io.IOBase):
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'fileobj' to be a file object")
        if filename is not None and type(filename) is not str:
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'filename' to be a str")
        elif not os.access(filename, os.R_OK):
            raise ValueError("kmerdb.graph.open expects an existing filepath as its first positional argument")

        elif mode is not None and type(mode) is not str:
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'mode' to be a str")
        elif n1_dtype is not None and type(n1_dtype) is not str:
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'n1_dtype' to be a str")
        elif n2_dtype is not None and type(n2_dtype) is not str:
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'n2_dtype' to be a str")
        elif weights_dtype is not None and type(weights_dtype) is not str:
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'weights_dtype' to be a str")

        elif sort is not None and type(sort) is not bool:
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'sort' to be a bool")
        elif slurp is not None and type(slurp) is not bool:
            raise TypeError("kmerdb.graph.KDBGReader expects the keyword argument 'slurp' to be a bool")

        # if logger is None:
        #     raise ValueError("graph.KDBGReader expects the keyword argument 'logger' to be valid")
        
        """

        Handle fileobj or 
        """
        
        if fileobj:
            assert filename is None
            handle = fileobj
            assert "b" in handle.mode.lower()
        else:
            if "w" in mode.lower() or "a" in mode.lower():
                raise ValueError("Must use read mode (default), not write or append mode")
            handle = _open(filename, "rb")

        self._text = "b" not in mode.lower()
        if self._text:
            self._newline = "\n"
        else:
            self._newline = b"\n"
        self._handle       = handle
        self._filepath     = self._handle.name

        #       Placeholder
        self.max_cache     = 100
        
        self._buffers      = {}
        self._block_start_offset = None
        self._block_raw_length = None

        self.logger = logger
        self._loggable = logger is not None
        """

        np.array defs

        n1
        n2
        weights

        """
        
        self.n1         = None
        self.n1_dtype   = None
        self.n2         = None
        self.n2_dtype   = None
        self.weights       = None
        self.weights_dtype = None

        self.read_and_validate_kdbg_header()

        if slurp is True:
            if self._loggable:
                self.logger.log_it("Reading .kdbg contents into memory", "INFO")
            self.slurp(n1_dtype=n1_dtype, n2_dtype=n2_dtype, weights_dtype=weights_dtype)

        self.is_int = True
        if handle is not None:
            handle.close()
        if self._handle is not None:
            self._handle.close()
        self._handle = None
        handle = None
        fileobj=None
        return

    def read_and_validate_kdbg_header(self):

        """

        KDBGReader PyYAML                metadata ingestion

        """

        '''
        Here we want to load the metadata blocks. We want to load the first two lines of the file: the first line is the version, followed by the number of metadata blocks
        '''
        # 0th block
        if self._loggable:
            self.logger.log_it("Loading the 0th block from '{0}'...".format(self._filepath), "INFO")
        self._load_block(self._handle.tell())

        self._buffer = self._buffer.rstrip(config.header_delimiter)
        
        initial_header_data = OrderedDict(yaml.safe_load(self._buffer))

        # Placeholder
        num_header_blocks = None

        if type(initial_header_data) is str:
            if self._loggable:
                self.logger.log_it("Inappropriate type for the header data.", "ERROR")
            #logger.info("Um, what the heckin' is this in my metadata block?")
            raise TypeError("kmerdb.graph.KDBGReader could not parse the YAML formatted metadata in the first blocks of the file")
        elif type(initial_header_data) is OrderedDict and self._loggable:
            self.logger.log_it("Successfully parsed the 0th block of the file, which is expected to be the first block of YAML formatted metadata", "INFO")
            self.logger.log_it("Assuming YAML blocks until delimiter reached.", "INFO")

        """

        KDBGReader
        
        YAML metadata spec validation
        """
            
        if "version" not in initial_header_data.keys():
            raise TypeError("kmerdb.graph.KDBGReader couldn't validate the header YAML")
        elif "metadata_blocks" not in initial_header_data.keys():
            raise TypeError("kmerdb.graph.KDBGReader couldn't validate the header YAML")

        if self._loggable:
            self.logger.log_it(initial_header_data, "INFO")
        
        if initial_header_data["metadata_blocks"] != 1:
            if self._loggable:
                self.logger.log_it("More than 1 metadata block: uhhhhh are we loading any additional blocks", "ERROR")
            raise IOError("Cannot read more than 1 metadata block yet")

        for i in range(initial_header_data["metadata_blocks"] - 1):
            if self._loggable:
                self.logger.log_it("Multiple metadata blocks read, most likely from a composite edge-graph...", "WARNING")
            self._load_block(self._handle.tell())
            addtl_header_data = yaml.safe_load(self._buffer.rstrip(config.header_delimiter))
            if type(addtl_header_data) is str:
                if self._loggable:
                    self.logger.log_it(str(addtl_header_data), "ERROR")
                raise TypeError("kmerdb.graph.KDBGReader determined the data in the {0} block of the header data from '{1}' was not YAML formatted".format(i, self._filepath))
            elif type(addtl_header_data) is dict:
                sys.stderr.write("\r")
                sys.stderr.write("Successfully parsed {0} blocks of YAML formatted metadata".format(i))
                initial_header_data.update(addtl_header_data)
                num_header_blocks = i
            else:
                if self._loggable:
                    self.logger.log_it(str(addtl_header_data), "ERROR")
                raise RuntimeError("kmerdb.graph.KDBGReader encountered a addtl_header_data type that wasn't expected when parsing the {0} block from the .kdb file '{1}'.".format(i, self._filepath))

        #raise RuntimeError("kmerdb.graph.KDBGReader encountered an unexpected type for the header_dict read from the .kdb header blocks")

        if self._loggable:
            self.logger.log_it("Validating the header YAML...", "INFO")

        try:
            jsonschema.validate(instance=initial_header_data, schema=config.graph_schema)
            self.metadata = dict(initial_header_data)
            
            self.k = self.metadata['k']
            self.n1_dtype = self.metadata["n1_dtype"]
            self.n2_dtype = self.metadata["n2_dtype"]
            self.weights_dtype = self.metadata["weights_dtype"]
            self.sorted = self.metadata["sorted"]

        except jsonschema.ValidationError as e:
            if self._loggable:
                self.logger.log_it("kmerdb.graph.KDBGReader couldn't validate the header/metadata YAML from {0} header blocks".format(num_header_blocks), "ERROR")

                self.logger.log_it("""


Failed to validate YAML header.

-------------------------------



 You can store unique k-mer counts, total nullomer counts, and other metadata alongside the weighted edge list with the 'kmerdb graph' command


 ...then try again to view the header with 'kmerdb header'



""", "ERROR")

                
                self.logger.log_it(e.__str__, "ERROR")
            raise ValueError("Requires kmerdb v{0} format YAML header. Body is .tsv format table, .bgzf file.       - weighted edge list        (idx, node1, node2, weight)")
        self.metadata["header_offset"] = self._handle.tell()
        #logger.debug("Handle set to {0} after reading header, saving as handle offset".format(self.metadata["header_offset"]))
        #self._reader = gzip.open(self._filepath, 'r')
        self._offsets = deque()
        for values in bgzf.BgzfBlocks(self._handle):
            #logger.debug("Raw start %i, raw length %i, data start %i, data length %i" % values)
            self._offsets.appendleft(values) # raw start, raw length, data start, data length
        if len(self._offsets) == 0:
            raise IOError("kmerdb.graph.KDBGReader opened an empty file")
        # Skip the zeroth block
        self._load_block()
        # print(str(self._buffer)) # 1
        # print(self.readline())
        # self._load_block()
        # print(self._buffer) # 2
        # print(self.readline())


    def read_line(self):
        """
        Read and parse a single line from the .kdbg file

        :returns: node1_id, node2_id, weight
        :rtype: tuple
        
        """
        
        line = self.readline()


        if self.n1_dtype == "uint64" and self.n2_dtype == "uint64" and self.weights_dtype == "uint64":
            return parse_kdbg_table_line(line, row_dtype="uint64")
        else:
            raise IOError("kmerdb.graph.KDBGReader.read_line Cannot determine file dtype")



    def slurp(self, n1_dtype:str="uint64", n2_dtype:str="uint64", weights_dtype:str="uint64"):
        """
        Autoload the .kdbg file into memory

        :param n1_dtype: NumPy dtype
        :type str:
        :param n2_dtype: NumPy dtype
        :type str:
        :param weights_dtype: NumPy dtype
        :type str:

        """
        
        if type(n1_dtype) is not str:
            raise TypeError("kmerdb.graph.KDBGReader.slurp expects n1_dtype to be a str")
        if type(n2_dtype) is not str:
            raise TypeError("kmerdb.graph.KDBGReader.slurp expects n2_dtype to be a str")
        if type(weights_dtype) is not str:
            raise TypeError("kmerdb.graph.KDBGReader.slurp expects weights_dtype to be a str")

        try:
            np.dtype(n1_dtype)
            np.dtype(n2_dtype)
            np.dtype(weights_dtype)
        except TypeError as e:
            raise TypeError("kmerdb.graph.KDBGReader.slurp expects each dtype keyword argument to be a valid numpy data type")

        vmem = psutil.virtual_memory()

        i = 0

        idx = []
        arrs = []
        node1 = []
        node2 = []
        weights = []
        reading=True
        while reading is True:

            try:
                line = next(self)

            except StopIteration as e:
                if self._loggable:
                    self.logger.log_it("Finished loading .kdbg through slurp (on init)", "ERROR")
                #raise e
                pass
            if line is None:
                if self._loggable:
                    self.logger.log_it("'next' returned None. Panic", "WARNING")

                raise RuntimeError("kmerdb.graph.KDBGReader.slurp Panicked on new line")


                  
            try:
                
                arr = self.read_line()
                arrs.append(arr)

                
                i, n1, n2, weight = arr

                idx.append(i)
                node1.append(n1)
                node2.append(n2)
                weights.append(weight)
            except StopIteration as e:
                reading = False

        
        self.n1 = np.array(node1, dtype=n1_dtype)
        self.n2 = np.array(node2, dtype=n2_dtype)
        self.weights = np.array(weights, dtype=weights_dtype)
        return



    
class KDBGWriter(bgzf.BgzfWriter):
    """
    A wrapper class around Bio.bgzf.BgzfWriter to write a .kdbg file to disk.


    :ivar filename: valid filepath
    :ivar mode: read/write mode
    :ivar metadata: header information
    :ivar both_strands: *unimplemented*
    :ivar fileobj: io.IOBase
    :ivar compresslevel: compression parameter
    """
    
    def __init__(self, filename:str=None, mode:str="w", metadata:OrderedDict=None, both_strands:bool=False, fileobj:io.IOBase=None, compresslevel:int=6, logger=None):
        """
        A wrapper around Bio.bgzf.BgzfWriter

        :param filename: A valid filepath
        :type filename: str
        :param mode: read/write mode
        :type mode: str
        :param metadata: A metadata header for the .kdbg file
        :type metadata: collections.OrderedDict
        :param both_strands: *unimplemented*
        :type both_strands: bool
        :param compresslevel: compression parameter
        :type compresslevel: int
        :raise TypeError: filename was not a str
        :raise TypeError: mode was invalid
        :raise TypeError: both_strands was invalid
        :raise TypeError: fileobj was invalid
        :raise TypeError: compresslevel was invalid
        :raise ValueError: mode was invalid
        :raise NotImplementedError: append mode was invalid
        """
        
        if filename is None or type(filename) is not str:
            raise TypeError("kmerdb.graph.KDBGWriter expects the filename to be a str")
        elif mode is None or type(mode) is not str:
            raise TypeError("kmerdb.graph.KDBGWriter expects the mode to be a str")
        elif metadata is None or (type(metadata) is not OrderedDict and type(metadata) is not dict):
            raise TypeError("kmerdb.graph.KDBGWriter - invalid metadata argument")
        elif both_strands is None or type(both_strands) is not bool:
            raise TypeError("kmerdb.graph.KDBGWriter expects the keyword argument 'both_strands' to be a bool")
        elif fileobj is not None and not isinstance(fileobj, io.IOBase):
            raise TypeError("kmerdb.graph.KDBGWriter expects the keyword argument 'fileobj' to be an instance of io.IOBase")
        elif compresslevel is None or type(compresslevel) is not int:
            raise TypeError("kmerdb.graph.KDBGWriter expects the keyword argument 'compresslevel' to be an int")

        self.logger = logger
        self._loggable = logger is not None

        
        if fileobj:
            assert filename is None, "kmerdb.graph expects filename to be None is fileobj handle is provided"
            handle = fileobj
        else:
            if "w" not in mode.lower() and "a" not in mode.lower():
                raise ValueError("Must use write or append mode, not %r" % mode)
            elif "wb" == mode:
                pass
            elif mode == "w":
                pass
            elif "a" in mode.lower():
                raise NotImplementedError("Append mode is not implemented yet")
                # handle = _open(filename, "ab")
            else:
                raise RuntimeError("Unknown mode for .kdbg file writing class kmerdb.graph.KDBGWriter: {0}".format(mode))
        self._text = "b" not in mode.lower()
        self._handle = _open(filename, "wb")
        self._buffer = b"" if "b" in mode.lower() else ""
        self.compresslevel = compresslevel


        self.logger = logger
        self._loggable = logger is not None
        
        """
        Write the header to the file
        """

        if self._loggable:
            self.logger.log_it("Constructing a new .kdbg file '{0}'...".format(self._handle.name), "INFO")


        # 3-04-2024
        yaml.add_representer(OrderedDict, util.represent_yaml_from_collections_dot_OrderedDict)

        self.metadata = metadata

        #self._write_block(metadata_slice)
        if "b" in mode.lower():
            metadata_bytes = bytes(yaml.dump(self.metadata, sort_keys=False), 'utf-8')
            metadata_plus_delimiter_in_bytes = metadata_bytes + bytes(config.header_delimiter, 'utf-8')
            self.metadata["metadata_blocks"] = math.ceil( sys.getsizeof(metadata_plus_delimiter_in_bytes) / ( 2**16 ) ) # First estimate
            metadata_bytes = bytes(yaml.dump(self.metadata, sort_keys=False), 'utf-8')
            metadata_bytes = metadata_bytes + bytes(config.header_delimiter, 'utf-8')
            self.metadata["metadata_blocks"] = math.ceil( sys.getsizeof(metadata_bytes) / ( 2**16 ) ) # Second estimate
            metadata_bytes = bytes(yaml.dump(self.metadata, sort_keys=False), 'utf-8')
            metadata_bytes = metadata_bytes + bytes(config.header_delimiter, 'utf-8')
            if self._loggable:
                self.logger.log_it("Writing the {0} metadata blocks to the new file".format(self.metadata["metadata_blocks"]), "INFO")
                self.logger.log_it("Header is being written as follows:\n{0}".format(yaml.dump(self.metadata, sort_keys=False)), "DEBUG")

            # 01-01-2022 This is still not a completely functional method to write data to bgzf through the Bio.bgzf.BgzfWriter class included in BioPython
            # I've needed to implement a basic block_writer, maintaining compatibility with the Biopython bgzf submodule.
            #self.write(bytes(yaml.dump(metadata, sort_keys=False), 'utf-8'))
        
            for i in range(self.metadata["metadata_blocks"]):
                metadata_slice = metadata_bytes[:65536]
                metadata_bytes = metadata_bytes[65536:]
                self._write_block(metadata_slice)

                #self._write_block
                self._buffer = b""
                self._handle.flush()
        elif "w" == mode.lower() or "x" == mode.lower():
            metadata_yaml_str = yaml.dump(metadata, sort_keys=False)
            metadata_bytes = bytes(metadata_yaml_str, 'utf-8')
            mode = "b"
            self._buffer = b""
            self.write(metadata_bytes)

            self._handle.flush()
        else:
            raise RuntimeError("Could not determine proper encoding for write operations to .kdb file")
        

    

        

class Parseable:
    def __init__(self, arguments, logger):
        self.arguments = arguments
        self.logger = logger
        
    def parsefile(self, filename):
        """Wrapper function for graph.parsefile to keep arguments succinct for deployment through multiprocessing.Pool
            
        :param filename: the filepath of the fasta(.gz)/fastq(.gz) to process with kmerdb.graph.parsefile
        :type filename: str
        """
        return parsefile(filename, self.arguments.k, quiet=self.arguments.quiet, logger=self.logger)



